/* $Id$
 *
 * dChecker - Distributed Behavior Protocols Checker 
 * Copyright (C) 2006-2007  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.ow2.dsrg.fm.dchecker;


import org.ow2.dsrg.fm.dchecker.architecture.DChPreprocessedArchitecture;
import org.ow2.dsrg.fm.dchecker.distributed.*;
import org.ow2.dsrg.fm.dchecker.distributed.spacemap.BoundingBox;
import org.ow2.dsrg.fm.dchecker.errors.*;



import java.io.*;
import java.util.List;

import org.ow2.dsrg.fm.bptools.BPEventTableImpl;
import org.ow2.dsrg.fm.bptools.IEventTable;
import org.ow2.dsrg.fm.bptools.Preferences;
import org.ow2.dsrg.fm.bptools.util.DOTPrintStream;
import org.ow2.dsrg.fm.bptools.util.Log;
import org.ow2.dsrg.fm.dchecker.fly.FlyFSM;
import org.ow2.dsrg.fm.dchecker.fly.FlyState;


/**
 * Main class of local version of checker
 */
public class LocalChecker {
    
    private static void printHelp() {
        System.out.println(
                "Usage: LocalCheck [file.bp | file.archbp] \n");
    }
    
    /**
     * Run the local version of the checker.
     *
     * The first argument is file containing architecture to be checked.
     * The second argument is file used to write dot output.
     *
     * The checker dumps out the architecture in dot format ( www.graphviz.org)
     * @param args list of arguments
     * @throws java.lang.Exception thrown when something goes wrong
     */
    public static void main(String[] args) throws Exception {
        Log.init();
        Reader s;
        
        if (args.length == 1) {
            if (args[0].equals("--help")) {
                printHelp();
                System.exit(1);
            }
        } else {
            printHelp();
            return;
        }
        
        IEventTable et= new BPEventTableImpl();
        DChPreprocessedArchitecture a = DChPreprocessedArchitecture.loadPreprocessedArchitecture(args[0],et);
        if ( a == null ) {
            printHelp();
            return;
        }
        
        new PrintStream(new FileOutputStream("preprocessedArch.bp")).print(a.print());
        
        
        DOTPrintStream dumpNFSMStream = null;
        DOTPrintStream dumpDFSMStream = null;
        
        String dumpParseTree = Preferences.getDumpParseTree();
        if (dumpParseTree!=null){
            DOTPrintStream o = new DOTPrintStream(dumpParseTree);
            a.dotDumpProtocols(o);
            o.DOTclose();
        }
        
        String dumpNFSM = Preferences.getDumpNFSM();
        if (dumpNFSM!=null){
            dumpNFSMStream = new DOTPrintStream(dumpNFSM);
        }
        
        String dumpDFSM = Preferences.getDumpDFSM();
        if (dumpDFSM!=null){
            dumpDFSMStream = new DOTPrintStream(dumpDFSM);
        }
        
        a.performTransformationChain(dumpNFSMStream,dumpDFSMStream);
        FlyFSM ffsm = a.getFlyfsm();
        
        String dumpDensity = Preferences.getDumpDensity();
        
        if (dumpDensity!=null){
            int [][] edgeDensity = ffsm.getEdgeDensity();
            for(int i = 0 ; i<edgeDensity.length; i++){
                FileWriter fw = new FileWriter(dumpDensity+"_"+i);
                int[] ed = edgeDensity[i];
                for (int j = 0; j<ed.length ; j++){
                    fw.write(String.valueOf(j)+"\t"+ed[j]+"\n");
                }
                fw.close();
            }
        }
        
        FlyState init = new FlyState();
        
        ffsm.init(init);
        
        BoundingBox bb = new BoundingBox(ffsm.getDimension());
        
        ffsm.fillBoundingBox(bb);
        
        ffsm.dump(new PrintStream("localdump"), et);
        
        Area area;
        if (Preferences.getTraversal()==Preferences.Traversal.DFS) {
            area = new DFSArea(bb);
        } else {
            area = new BFSArea(bb);
        }
        
        String dumpFname = Preferences.getDumpFile();
        if (dumpFname!=null){
            area.setUserHandler(new FileDump(dumpFname));
        }
        area.setFlyFSM(ffsm);
        area.newEntryPoint(init);
        
        int retcode = 0;
        try {
            while (area.traversal(null));
            
            System.out.println(
                    "Checked. " + area.getNumberOfVisited() + " states visited");
        } catch (NondeterministicConsentException e) {
            e.printStackTrace();
            System.out.println(
                    "Non-deterministic consent: "
                    + a.getEventTable().getEventString(e.event));
            retcode = 1;
        } catch (BadActivityException e) {
            e.printStackTrace();

            System.out.println("Event " + a.getEventTable().getEventString(e.event));

            System.out.println(a.getAnnotatedProtocol(e));
            List<FlyState> stackTrace = area.getStackTrace(e.f);
            System.out.println("CallStack:\n" + a.printCallStack(stackTrace));
            retcode = 1;
        } catch (CompositionException e) {
            e.printStackTrace();
            System.out.println(a.getAnnotatedProtocol(e));
            List<FlyState> stackTrace = area.getStackTrace(e.f);
            System.out.println("CallStack:\n" + a.printCallStack(stackTrace));
            retcode = 1;
        } catch (Exception e) {
            e.printStackTrace();
            retcode = 1;
        }
        
        area.finishUserHandler();
        System.exit(retcode);
    }
}
