/* $Id$
 *
 * dChecker - Distributed Behavior Protocols Checker 
 * Copyright (C) 2006-2007  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.ow2.dsrg.fm.dchecker.architecture;


import org.ow2.dsrg.fm.dchecker.*;
import org.ow2.dsrg.fm.dchecker.errors.*;

import java.util.*;
import java.util.logging.Logger;

import org.ow2.dsrg.fm.bptools.IEventTable;
import org.ow2.dsrg.fm.bptools.Preferences;
import org.ow2.dsrg.fm.bptools.ProtocolKeeper;
import org.ow2.dsrg.fm.bptools.architecture.BPParserBridge;
import org.ow2.dsrg.fm.bptools.architecture.ParsedArchitecture;
import org.ow2.dsrg.fm.bptools.architecture.PreprocessedArchitecture;
import org.ow2.dsrg.fm.bptools.fsm.CompactFSM;
import org.ow2.dsrg.fm.bptools.parser.nodes.BPComponent;
import org.ow2.dsrg.fm.bptools.parser.nodes.BPNode;
import org.ow2.dsrg.fm.bptools.parser.traversal.CollectEventNamesActions;
import org.ow2.dsrg.fm.bptools.parser.traversal.ParseTreeDOTActions;
import org.ow2.dsrg.fm.bptools.traversal.ActionPack;
import org.ow2.dsrg.fm.bptools.traversal.DFSPostfixTraversal;
import org.ow2.dsrg.fm.bptools.traversal.FSMDOTTraversal;
import org.ow2.dsrg.fm.bptools.traversal.Traversal;
import org.ow2.dsrg.fm.bptools.util.DOTPrintStream;

import org.ow2.dsrg.fm.dchecker.fly.BitPartInfo;
import org.ow2.dsrg.fm.dchecker.fly.FlyBinaryConsent;
import org.ow2.dsrg.fm.dchecker.fly.FlyFSM;
import org.ow2.dsrg.fm.dchecker.fly.FlyMultiConsent;
import org.ow2.dsrg.fm.dchecker.fly.FlyState;
import org.ow2.dsrg.fm.dchecker.fly.HierarchicalFlyFSM;
import org.ow2.dsrg.fm.dchecker.fly.SimpleFlyFSM;

/**
 * Container for preprocessed architecture.
 * PreprocessedArchitecture contains behavior specification of an architecture
 * with "instantiated" behavior protocols and low level bindings of methods.
 * Methods in the protocols are renamed acoording to bindings among components.
 * The protocols are in a form of behavior protocols on a formal level, so that
 * they are treated as if connected by the consent operator. Therefore, there
 * is also information about method-to-method bindings, unbound methods, and
 * multi-synchronization.
 *
 * @author Tomas Poch
 */
public class DChPreprocessedArchitecture extends PreprocessedArchitecture{
    
    
    /** FSM representation used to generate result state space on the fly. */
    private FlyFSM flyfsm;
    
    
    /**
     * Creates a new initially empty instance of Architecture.
     */
    
    public DChPreprocessedArchitecture(IEventTable iet) {
        super(iet);
    }
    
    public DChPreprocessedArchitecture(ParsedArchitecture pa, IEventTable iet){
        super(pa,iet);
    }
    
    /**
     * Create inverted copy of given parse tree. The inverted copy is copy of
     * the original parse tree with complementary leaves (!a -> ?a and vice
     * versa)
     *
     * @param p root of original parse tree
     * @return inverted copy of original parse tree
     
    public static BPNode invertProtocol(BPNode p, IEventTable et) {
        BPNode ret;
        ActionPack<BPNode, BPNode> a = new InvertProtocolActions(et);
        
        ret = a.action(p);
        return ret;
    }
    */



    
    /**
     * Connect Simple Finite State Machines by consent operator and create tree
     * of consent operators and Simple FSM that represents the architecture state space
     */
    public void makeFlyFSM() {
        SimpleFlyFSM[] s = new SimpleFlyFSM[components.size()];
        
        for (int i = 0; i < s.length; i++) {
            CompactFSM cfsm = components.get(i).getProtocolKeeper().getCompactFSM();
            s[i] = new SimpleFlyFSM(cfsm, unbound);
        }
        
        flyfsm = createConsent(s);
    }
    
    /**
     * Create consent operator over the given protocols automatons.
     *
     * @param simpleFSMs array of primitive automatons
     * @return FSM representing the consent
     */
    protected FlyFSM createConsent(SimpleFlyFSM[] simpleFSMs) {
        if (Preferences.getConsent() == Preferences.Consent.Binary) {
            HierarchicalFlyFSM hffsm = simpleFSMs[simpleFSMs.length - 1];
            
            for (int i = simpleFSMs.length - 2; i >= 0; i--) {
                hffsm = new FlyBinaryConsent(simpleFSMs[i], hffsm,
                        bindings.get(i), unbound);
            }
            hffsm.setBase(BitPartInfo.ZERO);
            return  hffsm;
        } else {
            return new FlyMultiConsent(simpleFSMs, bindings, unbound);
        }
    }
    
    /**
     * Create consent operator over the protocol automatons of this architecture.
     *
     * The primitive automatons are truncated so that the resultant automaton
     * contains only the state identifiers that belongs to the N-dimensional cube
     * (where N is a number of protocols in the architecture) defined by the input arrays
     *
     * @param min minimal identifier that belongs into cube
     * @param max maximal identifier that belongs into cube
     *
     * @return FSM representing truncated consent
     */
    
    public FlyFSM getFlyFSMSlice(int[] min, int[] max) {
        SimpleFlyFSM[] slices = new SimpleFlyFSM[components.size()];
        
        for (int i = 0; i < slices.length; i++) {
            CompactFSM cfsm = components.get(i).getProtocolKeeper().getCompactFSM();
            slices[i] = new SimpleFlyFSM(cfsm, unbound).getSlice(min[i],
                    max[i]);
            
        }
        return createConsent(slices);
    }
    
    /**
     * Take all parse trees and create corresponding automatons.
     * The Deterministic version is skipped if it is possible to
     * create Simple FSM from Non-deterministic
     * @param nfsmStream stream to dump all non-deterministic fsms
     * @param dfsmStream stream to dump all deterministic fsms
     */
    public void performTransformationChain(DOTPrintStream nfsmStream,
            DOTPrintStream dfsmStream) {
        super.performTransformationChain(nfsmStream, dfsmStream);
        
        makeFlyFSM();
    }
    
    /**
     * Create textual representation of behavior protocol. The position marked with @
     * symbol is where the error occured.
     * @param e exception thrown by traversal when composition error occure
     * @return annotated protocol
     */
    public String getAnnotatedProtocol(CompositionException e) {
        if (!Preferences.annotatedProtocol()) {
            return "Annotated protocols disabled. Use -DannotatedProtocol=yes";
        }
        
        StringBuffer sb = new StringBuffer();
        int[] splitted = new int[flyfsm.getDimension()];
        
        flyfsm.getStateParts(e.f, splitted);
        
        for (int i = 0; i < components.size(); i++) {
            sb.append(
                    "Protocol " + i + ":\n"
                    + components.get(i).getProtocolKeeper().getAnnotatedProtocol(splitted[i],
                    eventTable)
                    + "\n");
        }
        return sb.toString();
    }
    
    final static int NOT_YET_ASSIGNED = -1;
    
    /**
     * Create string representation of the given call stack.
     * This method is used to obtain human readable trace.
     * @param lst list of states.
     * @return string callstack
     */
    public String printCallStack(List<FlyState> lst) {
        
        StringBuffer ret = new StringBuffer();
        final int dim = flyfsm.getDimension();
        int[] splittedLast = new int[dim];
        int[] splittedNext = new int[dim];
        int[] xchg;
        Iterator<FlyState> it = lst.iterator();
        FlyState next;
        
        if (it.hasNext()) {
            next = it.next();
            flyfsm.getStateParts(next, splittedNext);
        }
        while (it.hasNext()) {
            next = it.next();
            xchg = splittedLast;
            splittedLast = splittedNext;
            splittedNext = xchg;
            flyfsm.getStateParts(next, splittedNext);
            int event1 = NOT_YET_ASSIGNED;
            int event2 = NOT_YET_ASSIGNED;
            
            for (int i = 0; i < dim; i++) {
                int lastStateOfIthAutomaton = splittedLast[i];
                int nextStateOfIthAutomaton = splittedNext[i];
                
                if (lastStateOfIthAutomaton != nextStateOfIthAutomaton) {
                    int edgeLabel = components.get(i).getProtocolKeeper().getCompactFSM().getEdgeLabelConnecting(
                            lastStateOfIthAutomaton, nextStateOfIthAutomaton);
                    
                    if (event1 == NOT_YET_ASSIGNED) {
                        event1 = edgeLabel;
                    } else if (event2 == NOT_YET_ASSIGNED) {
                        event2 = edgeLabel;
                    } else {
                        assert false;
                    }
                }
            }
            
            stateId(splittedLast, ret);
            ret.append('\t');
            
            if (event2 == NOT_YET_ASSIGNED) {
                if (event1 == NOT_YET_ASSIGNED) {
                    assert false;
                }
                ret.append(eventTable.getEventString(event1));
            } else {
                ret.append(eventTable.getEventString(event1) + eventTable.getEventString(event2));
            }
            
            ret.append('\n');
            
        }
        
        stateId(splittedNext, ret);
        ret.append('\n');
        
        return ret.toString();
        
    }
    
    /**
     *  Create string representation of the state
     *  @param splitted state identifier splitted into array
     *  @param sb result is stored here
     */
    public static void stateId(int[] splitted, StringBuffer sb) {
        sb.append('[');
        for (int i = 0; i < splitted.length; i++) {
            sb.append(splitted[i]);
            if (i != splitted.length - 1) {
                sb.append(',');
            }
        }
        sb.append(']');
        
    }
    
    /**
     * Dump the protocols parse trees to dot file ( www.graphviz.org)
     * @param ps output dot file
     */
    public void dotDumpProtocols(DOTPrintStream ps) {
        ActionPack a = new ParseTreeDOTActions(ps, eventTable);
        Traversal t = new DFSPostfixTraversal(a);
        
        int i = 0;
        
        for (BPComponent c: components) {

            BPNode bp = c.getBehavior();
            
            try {
                t.traverse(bp);
            } catch (Exception e) {
                // there should not be any assertion in the ParseTreeDOTActions
                assert false;
            }
            
            if (i < bindings.size()) {
                String label = null;
                
                for (Integer event:bindings.get(i)) {
                    if (label == null) {
                        label = eventTable.getEventString(event);
                    } else {
                        label = label + "," + eventTable.getEventString(event);
                    }
                }
                
                ps.node(bindings.get(i).hashCode(), label);
                
            }
            i++;
            
            ps.print("\n\n\n");
        }
    }
    
    /**
     * Dump the non-deterministic automatons to dot file ( www.graphviz.org)
     * @param ps output dot file
     */
    public void dotDumpNDTFSMS(DOTPrintStream ps) {
        FSMDOTTraversal t = new FSMDOTTraversal(ps, eventTable);
        
        for (BPComponent c: components) {
            ProtocolKeeper pt = c.getProtocolKeeper();
            try {
                pt.dotDumpNDTFSMS(t);
            } catch (Exception e) {
                // there should not be any assertion in the FSMDOTActions
                assert false;
            }
            ps.print("\n\n\n");
            
        }
        
    }
    
    /**
     * Dump the deterministic automatons to dot file ( www.graphviz.org)
     * If some of automaton is not created (because the non-deterministic version
     * was already deterministic ) the non-deterministic version is dumped instead
     * of it.
     * @param ps output dot file
     */
    public void dotDumpDTFSMS(DOTPrintStream ps) {
        FSMDOTTraversal t = new FSMDOTTraversal(ps, eventTable);
        
        for (BPComponent c:components) {
            ProtocolKeeper pt = c.getProtocolKeeper();
            try {
                pt.dotDumpDTFSMS(t);
            } catch (Exception e) {
                // there should not be any assertion in the FSMDOTActions
                assert false;
            }
            ps.print("\n\n\n");
            
        }
        
    }
    
    /**
     * Static method used to create new Architecture. The initial architectire
     * description - parse tree of behaviour protocol is parsed from the input stream
     * @param stream input stream
     * @return new architecture
     * @throws bpchecker.parser.javacc.ParseException thrown if syntax error is found
     
    public static PreprocessedArchitecture loadPreprocessedArchitecture(
            InputStream stream) throws ParseException {
        
        BPParser p = new BPParser(stream);
        PreprocessedArchitecture a = new PreprocessedArchitecture();
        
        p.architecture(a);
        a.logger().info("Number of events: " + a.eventTable.numberOfEvents());
        a.printEventInfo();
        return a;
    }
    
    /**
     * Static method used to create new Architecture. The initial architectire
     * description - parse tree of behaviour protocol is parsed from the given reader
     * @param reader input reader
     * @throws bpchecker.parser.javacc.ParseException thrown if syntax error is found
     * @return new architecture
     
    public static DChPreprocessedArchitecture loadPreprocessedArchitecture(
            java.io.Reader reader) throws ParseException {
        BPParser p = new BPParser(reader);
        DChPreprocessedArchitecture a = new DChPreprocessedArchitecture(new EventTableImpl());
        p.architecture(a);
        a.printEventInfo();
        return a;
    }
    */
    
    public static DChPreprocessedArchitecture loadPreprocessedArchitecture(
            String fName, IEventTable ie) throws Exception {
        ParsedArchitecture pa = ParsedArchitecture.loadArchitecture(fName,new BPParserBridge());
        return new DChPreprocessedArchitecture(pa,ie);
    }
    
    
    public void printEventSet(StringBuffer sb,Collection<Integer> s){
        TreeSet<String> al = new TreeSet<String>();
        for (Integer e:s){
            al.add(eventTable.getEventString(e));
        }
        
        for(String ename:al){
            sb.append(ename);
            sb.append('\n');
        }
        
    }
    
    public void printEventInfo(){
        CollectEventNamesActions cena = new CollectEventNamesActions();
        for (BPComponent c: components) {
            ProtocolKeeper pk = c.getProtocolKeeper();
            pk.getParseTree().apply(cena);
        }
        
        StringBuffer sb = new StringBuffer("\nMethods from protocols\n");
        printEventSet(sb,cena.events);
        
        HashSet<Integer> eventsfromSyncSet = new HashSet<Integer>();
        for (Set<Integer> si:bindings){
            eventsfromSyncSet.addAll(si);
        }
        if (unbound!=null){
            eventsfromSyncSet.addAll(unbound);
        }
        sb.append("\nMethods from sync sets and unbound\n");
        printEventSet(sb,eventsfromSyncSet);
        
        cena.events.removeAll(eventsfromSyncSet);
        sb.append("\nDifference\n");
        printEventSet(sb,cena.events);
        
        logger().info(sb.toString());
        
    }
    
    /**
     * Obtain event table created during initial parsing. The Event table contains item
     * for each event that occure in architecture description.
     * @return event table
     */
    public IEventTable getEventTable() {
        return eventTable;
    }
    
    
    public String print(){
        StringBuilder sb = new StringBuilder();
        for (BPComponent c: components) {
            sb.append("component{\n");
            sb.append(c.print(eventTable));
            sb.append("\n}\n");
        }
        return sb.toString();
    }
    
    /**
     * Get result of transformation chain. The given FlyFsm is capable to generate state
     * space of cartesian product of in-memory automatons connected by consent operator
     * on the fly.
     * @return automaton representatnion used for (distributed) traversal
     */
    public FlyFSM getFlyfsm() {
        return flyfsm;
    }
    
    /**
     * Get unbound events
     * @return list of unbound events
     */
    public Set<Integer> getUnbound() {
        return unbound;
    }
    
    /**
     * Getter for syncs.
     * @return List of syncs
     */
    public List<Set<Integer>> getSyncs() {
        return bindings;
    }
    
    /**
     * Getter for protocols
     * @return List of protocols
     */
    public List<BPComponent> getComponents() {
        return components;
    }
    
    /**
     * Obtain a logger
     * @return logger
     */
    protected Logger logger() {
        return Logger.getLogger(
                "dchecker.architecture.PreprocessedArchitecture");
    }
}
