/* $Id$
 *
 * dChecker - Distributed Behavior Protocols Checker 
 * Copyright (C) 2006-2007  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.ow2.dsrg.fm.dchecker.distributed;


import org.ow2.dsrg.fm.dchecker.errors.CompositionException;
import org.ow2.dsrg.fm.dchecker.fly.FlyFSM;
import org.ow2.dsrg.fm.dchecker.rmi.RMIClientTaskImpl;
import java.util.List;
import java.util.Set;
import org.ow2.dsrg.fm.dchecker.distributed.spacemap.BoundingBox;
import org.ow2.dsrg.fm.dchecker.fly.FlyState;
import org.ow2.dsrg.fm.dchecker.rmi.MachineLostException;


/**
 * This class represents part of state space assigned to particular client.
 * The state space is n-dimensional ( cartesian product of n automatons) and
 * state identifier consist of concated identifiers of all these automatons.
 *
 * The area is rectangular , remembers multiple stacks  ( for each entry point
 * to the area ) and remembers already visited states.
 *
 * The Area is created by the server (task assignment to the first client ) or by
 * another client ( division of already assigned area ) and transfered to the target.
 * The target initializes the Area and runs the traversal.
 *
 * @author thorm
 *
 */

public abstract class Area extends BoundingBox {
    
    /**
     * Constructor
     *
     * The area position is given by the bounding box
     * @param si position and size of new area
     */
    public Area(BoundingBox si) {
        this(si.getMinimalCoord(), si.getMaximalCoord());
        
    }

    /**
     * The new area position and size is given by two corners
     * @param min coordinates of first corner
     * @param max coordinates of second corner
     */
    public Area(int[] min, int[] max) {
        super(min, max);
        splittedState = new int[min.length];
        splittedNewState = new int[min.length];
    }
    
    /**
     * Insert new state reached from other area or initial state among entry points
     * @param state state identifier
     */
    public abstract void newEntryPoint(FlyState state);

    /**
     * Run the traversal
     * @return true if at leas one new state was visited
     * @param task client-side task data used to report states out of this area
     * @throws bpchecker.errors.CompositionException composition error detected
     * @throws java.io.IOException sending state identifier to server failed
     */
    public abstract boolean traversal(RMIClientTaskImpl task) throws CompositionException, MachineLostException;
    
    /**
     * Split the area into two parts and return one of them
     * @return untouched part of area
     */
    public abstract Area getSubtask();
    
    /**
     * Set the FSM for this area. The FSM must contain at least all states that belongs
     * to this area
     * @param f used to generate state space
     */
    public void setFlyFSM(FlyFSM f) {
        ffsm = f;
        /*BoundingBox bb = new BoundingBox(minimal.length);
        f.fillBoundingBox(bb);
        System.out.println("FFSM " + bb.toString() +"added into " + toString());*/
        
    }
    
    public FlyFSM getFlyFSM() {
        return ffsm;
    }
    
    
    /**
     * Return true if given state is member ot this area.
     * @param s state
     * @return true if the state is member of this area
     */
    public boolean belongs(FlyState s) {
        ffsm.getStateParts(s, splittedState);
        return belongs(splittedState);
    }
    
    public void setClientName(String s){
        visitedLocal.message = s + ": " +visitedLocal.message;
        visitedOther.message = s + ": " +visitedOther.message;
    }

    /**
     * Get path beginning with one of entry points.
     * @param last last state of the path
     * @return path
     */
    public List<FlyState> getStackTrace(FlyState last) {
        
        List<FlyState> ret = visitedLocal.getStackTrace(last);

        if (ret.size() == 1) {
            FlyState first = ret.get(0);

            if (first.data == first.parentData) {
                return null;
            }
        }
        
        return ret;
    }
    
    /**
     * Return number of visited states
     * @return number or visited states
     */
    public int getNumberOfVisited() {
        return visitedLocal.size();
    }
    
    protected UserHandler userHandler;
    
    /**
     * array used to temporarily keep state identifiers splitted into parts
     */
    protected int[] splittedState;
    
    /**
     * array used to temporarily keep new state identifier splitted into parts
     */
    protected int[] splittedNewState;
    
    /**
     * Keeps already visited states to avoid infinity loops and redundant traversals
     */
    protected VisitedNodes visitedLocal;
    
    /**
     * 
     */
    protected VisitedNodes visitedOther;
    
    /**
     * Finite state machine capable to generate state space of architecture on the fly.
     * It consist from number of deterministic automatons connected by consent operator
     */
    protected FlyFSM ffsm;

    /**
     * List of unbound events.
     */
    protected Set<Integer> unbound;
    
    public void finishUserHandler(){
    if (userHandler!=null){
            userHandler.finish();
        }
    }
    
    public UserHandler getUserHandler() {
        return userHandler;
    }

    public void setUserHandler(UserHandler userHandler) {
        this.userHandler = userHandler;
        for (FlyState fs:visitedLocal.getVisitedStates()){
            ffsm.getStateParts(fs, splittedState);
            userHandler.node(splittedState);
            //assert belongs(splittedState);
        }
    }
}