/* $Id$
 *
 * dChecker - Distributed Behavior Protocols Checker 
 * Copyright (C) 2006-2007  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.ow2.dsrg.fm.dchecker.distributed;


import org.ow2.dsrg.fm.dchecker.fly.FlyState;
import org.ow2.dsrg.fm.dchecker.rmi.*;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.rmi.RemoteException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;


/**
 * Server keeps information about registered clients here.
 * @author thorm
 */
public class ClientRepository {
    
    /** Creates a new instance of ClientRepository */
    public ClientRepository() {
        
        clients = new HashMap<Integer, ClientInfo>();
        
    }
    
    /**
     * Register new client
     * It adds given RMI stub into list of clients and assigns an identifier to it.
     * @param c new client
     * @return assigned identifier
     */
    public  synchronized int registerClient(Client c,InetSocketAddress ia) {
        
        int ret = idCounter++;
        
        clients.put(ret, new ClientInfo(c,ia));
        notifyAll();
        return ret;
    }
    
    /**
     * Unregister client
     * @param clientId client to be unregistred
     */
    public  synchronized void unregisterClient(int clientId) throws IOException {
        RMIServerTaskImpl.ClientData data = clients.get(clientId).clientTask;
        
        clients.remove(clientId);
        notifyAll();
    }
    
    public int size() {
        return clients.size();
    }
    
    /**
     * Get RMI stub associated with given identifier
     * @param clientId client identifier
     * @return Client associated with given identifier
     */
    
    public synchronized Client getClient(int clientId) {
        ClientInfo info = clients.get(clientId);
        
        return (info == null) ? null : info.client;
    }
    
    /*
     public synchronized RMIServerTaskImpl.ClientData getClientTaskData(int clientId){
     return clients.get(clientId).taskData;
     }
     */
    
    public synchronized Area getSubtask(int clientId) throws ClientLostException {
        try {
            return clients.get(clientId).clientTask.clientTask.getSubtask();
        } catch (RemoteException e) {
            throw new ClientLostException(clientId);
        }
    }
    
    public synchronized void addArea(int clientId, Area a) throws ClientLostException {
        try {
            clients.get(clientId).clientTask.clientTask.addArea(a);
        } catch (RemoteException e) {
            throw new ClientLostException(clientId);
        }
    }
    
    public synchronized List<FlyState> getStackTrace(int clientId, FlyState last) throws ClientLostException {
        try {
            return clients.get(clientId).clientTask.clientTask.getStackTrace(last);
        } catch (RemoteException e) {
            throw new ClientLostException(clientId);
        }
    }
    /*
    public synchronized void sendEntryPoint(int clientId, FlyState entryPoint) throws ClientLostException {
            clients.get(clientId).clientTask.writeState(entryPoint);
    }
    */
    public synchronized void setClientTaskData(int clientId, RMIServerTaskImpl.ClientData d) {
        clients.get(clientId).clientTask = d;
    }
    
    public synchronized void cleanClientTaskData() {
        for (ClientInfo ci:clients.values()) {
            ci.clientTask = null;
        }
    }
    
    /**
     * Get upper bound of all client identifiers.
     * @return maximal assigned client identifier
     */
    public synchronized int getMaxClientId() {
        return idCounter;
    }
    
    
       
    /*
    public synchronized long terminationDetection() throws ClientLostException {
        //long ret = 0;
        long totalSent = 0 ;
        long totalReceived = 0;
        
        TerminationInfo fi;
        
        for (Entry<Integer, ClientInfo> ci:clients.entrySet()) {
            ClientInfo value = ci.getValue();
            
            try {
                RMIServerTaskImpl.ClientData taskData = value.clientTask;
                
                if (taskData != null) {
                    fi = taskData.terminationDetection();
                    if (fi==null) return -1;
                    
                    totalSent += fi.statesSent;
                    totalReceived += fi.statesReceived;
                    
                }
                
            } catch (RemoteException ce) {
                throw new ClientLostException(ci.getKey());
            } catch (IOException ioe) {
                throw new ClientLostException(ci.getKey());
            }
        }
        
        long ret = totalSent - totalReceived;
        assert ret >= 0: "Received more than sent: "+ret;
        return ret;
        
    }*/
    
    
    public synchronized long finish() throws ClientLostException{
        long ret=0;
        for (Entry<Integer, ClientInfo> ci:clients.entrySet()) {
            ClientInfo value = ci.getValue();
            try {
                RMIServerTaskImpl.ClientData taskData = value.clientTask;
                
                if (taskData != null) {
                    ret += taskData.finish();
                }
                
            } catch (RemoteException ce) {
                throw new ClientLostException(ci.getKey());
            } catch (IOException ioe) {
                throw new ClientLostException(ci.getKey());
            }
            
        }
        return ret;
    }
    
    public Map<Integer,InetSocketAddress> createClientAddressMap(){
        Map<Integer,InetSocketAddress> ret = new HashMap<Integer,InetSocketAddress>();
        for (Map.Entry<Integer,ClientInfo> e: clients.entrySet()){
            ret.put(e.getKey(),e.getValue().isa);
        }
        return ret;
    }
    
    public Set<Integer> getClientIds() {
        return clients.keySet();
    }
    
    public void pingThemAll() throws ClientLostException {
        for (Entry<Integer, ClientInfo> ci:clients.entrySet()) {
            ClientInfo value = ci.getValue();
            
            try {
                value.client.ping();
            } catch (RemoteException ce) {
                throw new ClientLostException(ci.getKey());
            }
        }
    }
    
    public void resumeThemAll() throws ClientLostException {
        for (Entry<Integer, ClientInfo> ci:clients.entrySet()) {
            ClientInfo value = ci.getValue();
            
            try {
                value.clientTask.clientTask.resume();
            } catch (RemoteException ce) {
                throw new ClientLostException(ci.getKey());
            }
        }
    }
    
    public void pauseThemAll() throws ClientLostException {
        for (Entry<Integer, ClientInfo> ci:clients.entrySet()) {
            ClientInfo value = ci.getValue();
            
            try {
                value.clientTask.clientTask.pause();
            } catch (RemoteException ce) {
                throw new ClientLostException(ci.getKey());
            }
        }
    }
    
    
    
    /**
     * List of registered clients stubs. The index is client identifier
     */
    protected HashMap<Integer, ClientInfo> clients;
    
    protected int idCounter;
    
    protected static class ClientInfo {
        public ClientInfo(Client c,InetSocketAddress ai) {
            client = c;
            isa = ai;
        }
        
        InetSocketAddress isa;
        Client client;
        RMIServerTaskImpl.ClientData clientTask;
        
    }
}
