/* $Id$
 *
 * dChecker - Distributed Behavior Protocols Checker 
 * Copyright (C) 2006-2007  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.ow2.dsrg.fm.dchecker.distributed;


import org.ow2.dsrg.fm.dchecker.rmi.RMIClientTaskImpl;
import java.io.Serializable;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.PriorityQueue;
import java.util.Stack;

import org.ow2.dsrg.fm.dchecker.distributed.spacemap.BoundingBox;
import org.ow2.dsrg.fm.dchecker.errors.CompositionException;
import org.ow2.dsrg.fm.dchecker.fly.FlyFSM;
import org.ow2.dsrg.fm.dchecker.fly.FlyState;
import org.ow2.dsrg.fm.dchecker.rmi.MachineLostException;

/**
 *
 * @author pocht1am
 */
public class DFSArea extends Area {
    
    /**
     * Constructor
     *
     * The area position is given by the bounding box
     * @param si position and size of new area
     */
    public DFSArea(BoundingBox si) {
        this(si.getMinimalCoord(), si.getMaximalCoord());
        
    }
    
    /**
     * The new area position and size is given by two corners
     * @param min coordinates of first corner
     * @param max coordinates of second corner
     */
    public DFSArea(int[] min, int[] max) {
        super(min, max);
        int dim = min.length;
        
        splittedState = new int[dim];
        splittedNewState = new int[dim];
        visitedWithBounds = new VisitedNodesWithUpperBound(toString()+" Visited: ", dim);
        visitedLocal = visitedWithBounds;
        visitedOther = new VisitedNodes(toString()+" Outgoing: ");
    }
    
    {
        pendingNewAreaRequests = new LinkedList<NewAreaRequest>();
        entryPoints = new PriorityQueue<AbsValueFlyState>(10);
        bfsVisited = new HashSet<FlyState>();
    }
    
    
    
    /**
     * Add new entry point into the area.
     * Adds a new stack containing given state into the list of stacks.
     * @param s new state
     */
    public synchronized void newEntryPoint(FlyState state) {
        AbsValueFlyState avfs;
        if (!visitedLocal.contains(state)){
            if (ffsm!=null){ //not an initial state//
                ffsm.getStateParts(state,splittedState);
                avfs = new AbsValueFlyState(state,splittedState);
            }else{
                avfs = new AbsValueFlyState(state);
            }
            entryPoints.add(avfs);
        }
        
        
    }
    
    
    
    /**
     * This implements DFS traversal of given areas using given entry points.
     * The traversal is bounded by areas boundaries and if it reaches some state
     * that does not belong to the area the server is notified.
     * @return true if at least one state was processed
     * @param server Server representation of current task
     * @throws java.rmi.RemoteException communication error when area boundary was crossed
     * @throws bpchecker.errors.CompositionException composition error found
     */
    public boolean traversal(RMIClientTaskImpl task) throws CompositionException, MachineLostException{
        boolean workDone = false;
        
        FlyState newState;
        FlyFSM.StateNeighbourEnumeration top;
        int nextSymbol;
        synchronized(this){
            active = true;
        }
        if (!entryPoints.isEmpty()) {
            FlyState entryPoint = entryPoints.poll().fs;
            stack = new Stack<FlyFSM.StateNeighbourEnumeration>();
            putNewStateIntoStack(task,entryPoint,null);
            
            while (!stack.empty() ) {
                workDone = true;
                processPendingNewAreaRequests();
                
                top = stack.peek();
                
                if (top.hasMoreElements()) {
                    newState = top.nextElement();
                    putNewStateIntoStack(task,newState,top.getState());
                    
                } else {
                    stack.pop();
                }
                
            }
            
        }
        synchronized (this) {
            active = false;
        }
        processPendingNewAreaRequests();
        return workDone;
    }
    
    
    protected void putNewStateIntoStack(RMIClientTaskImpl task,FlyState newState,FlyState parent) throws MachineLostException{
        if (!visitedLocal.contains(newState)) {
            ffsm.getStateParts(newState, splittedNewState);
            if (belongs(splittedNewState)) {
                stack.push(ffsm.getNeighboursEnum(newState));
                visitedWithBounds.insert(newState, parent,
                        splittedNewState);
                
            } else {
                if (visitedOther.insert(newState, parent)) {
                    task.outOfArea(newState, splittedNewState);
                }
                
            }
        }
    }
    
    
    void processPendingNewAreaRequests() {
        synchronized (pendingNewAreaRequests) {
            while (!pendingNewAreaRequests.isEmpty() ) {
                NewAreaRequest rq = pendingNewAreaRequests.poll();
                
                synchronized (this) {
                    rq.result = getSubtaskImpl();
                    rq.done = true;
                    notifyAll();
                }
            }
        }
    }
    
    private boolean isFarEnaugh(int[] splitted,int axis,int smallest) {
        return splitted[axis]>=smallest;
    }
    
    
    
    private FlyState getStateFarEnaugh(int axis,int min, int[] toStoreSplitted){
        LinkedList<FlyState> queue = new LinkedList<FlyState>();
        
        for(FlyFSM.StateNeighbourEnumeration sne:stack){
            FlyState fs = sne.getState();
            if (bfsVisited.contains(fs)) break;
            queue.add(fs);
        }
        FlyState ret = getStateFarEnaugh(axis,min,toStoreSplitted,queue);
        if (ret!=null){
            return ret;
        }
        
        for (AbsValueFlyState avfs:entryPoints){
            queue.add(avfs.fs);
        }
        return getStateFarEnaugh(axis,min,toStoreSplitted,queue);
        
    }
    
    
    
    
    
    private FlyState getStateFarEnaugh(int axis,int min ,int[] toStoreSplitted,LinkedList<FlyState> queue) {
        
        
        FlyState fs,child;
        
        while((fs=queue.poll())!=null){
            FlyFSM.StateNeighbourEnumeration neighEnum  = ffsm.getNeighboursEnum(fs);
            try{
                while (neighEnum.hasMoreElements()){
                    child = neighEnum.nextElement();
                    if (visitedLocal.contains(child)||!bfsVisited.add(child)){
                        continue;
                    }
                    ffsm.getStateParts(child,toStoreSplitted);
                    if (!belongs(toStoreSplitted)){
                        continue;
                    }
                    if (isFarEnaugh(toStoreSplitted,axis,min)){
                        return child;
                    }
                    queue.add(child);
                }
            }catch (CompositionException e){}
        }
        return null;
        
        
    }
    
    public Area getSubtaskImpl() {
        int dim = ffsm.getDimension();
        
        //if(getWidth(getNarrowestAxis())<10) return null;
        
        int axis = getWidestAxis();
        int actualBound = visitedWithBounds.getBound(axis);
        int min = actualBound+ ((maximal[axis] - actualBound)/2);
        
        int[] areaEntryPointSplitted = new int[dim];
        
        FlyState areaEntryPoint = getStateFarEnaugh(axis,min,areaEntryPointSplitted);
        
        if (areaEntryPoint == null) {
            return null;
        }
        
        bfsVisited.clear();
        
        DFSArea ret = new DFSArea(splitAt(axis,min));
        
        ret.newEntryPoint(areaEntryPoint);
        
        Iterator<AbsValueFlyState> i = entryPoints.iterator();
        
        while (i.hasNext()) {
            AbsValueFlyState avfs = i.next();
            FlyState fs = avfs.fs;
            
            if (!belongs(fs)) {
                ret.newEntryPoint(fs);
                i.remove();
            }
        }
        
        System.out.println("New area created. Volume:" + ret.getVolume());
        return ret;
    }
    
    public Area getSubtask() {
        int i = 0;
        synchronized (this) {
            if (active) {
                NewAreaRequest nrq = new NewAreaRequest();
                
                synchronized (pendingNewAreaRequests) {
                    pendingNewAreaRequests.add(nrq);
                }
                
                while (nrq.done == false) {
                    try {
                        wait();
                    } catch (InterruptedException e) {}
                }
                
                return nrq.result;
            } else {
                return getSubtaskImpl();
            }
        }
    }
    
    
    
    /**
     * List of the stacks to be processed
     */
    private Stack<FlyFSM.StateNeighbourEnumeration> stack;
    
    private LinkedList<NewAreaRequest> pendingNewAreaRequests;
    
    private static class NewAreaRequest implements Serializable{
        Area result;
        boolean done = false;
    }
    
    private HashSet<FlyState> bfsVisited;
    
    
    private boolean active;
    
    VisitedNodesWithUpperBound visitedWithBounds;
    
    PriorityQueue<AbsValueFlyState> entryPoints;
    
    private static class AbsValueFlyState implements Serializable,Comparable<AbsValueFlyState>{
        public AbsValueFlyState(FlyState f){
            fs = f;
            absValue = 0;
        }
        
        public AbsValueFlyState(FlyState f,int[] splitted){
            fs = f;
            absValue = 0;
            for (int part:splitted){
                if (part>absValue) {
                    absValue = part;
                }
            }
            assert absValue>=0;
        }
        
        public int compareTo(AbsValueFlyState avfs) {
            return (absValue-avfs.absValue>0)?1:-1;
        }
        
        
        
        FlyState fs;
        private int absValue;
    }
}
