/* $Id$
 *
 * dChecker - Distributed Behavior Protocols Checker 
 * Copyright (C) 2006-2007  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.ow2.dsrg.fm.dchecker.distributed;


import org.ow2.dsrg.fm.dchecker.fly.FlyState;
import java.io.Serializable;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;


/**
 * This class is used to store visited states. 
 * The state identifier consists of n numbers of different bit-sizes. 
 * The parent of the state is remembered within the state itself. Parent is 
 * the first state the state was reached from.
 * @author thorm
 */
public class VisitedNodes implements Serializable {
    private int sizeNotification = 100000;

    /**
     * Create a new instance of VisitedNodes
     */
    public VisitedNodes(String m) {
        message = m;
        visited = new HashMap<FlyState, FlyState>();
    
    }
    
    
    /**
     * Create a new instance of VisitedNodes
     */
    public VisitedNodes(String m, int sizeNotification) {
        message = m;
        visited = new HashMap<FlyState, FlyState>();
        this.sizeNotification = sizeNotification;
    }
    
    /**
     * Return number of remembered states
     * @return number of remembered states
     */
    public synchronized int size() {

        /* for(int i = 0 ; i<distribution.length; i++){
         System.out.println(Integer.toString(i)+":"+distribution[i]);
         }*/
        return visited.size();
    }
    
    /**
     * Remember a root of the tree. The root is the initial state or a state
     * that came from another area. In the second case it already contains 
     * data of its parent identifier which is remembered by the another area
     *
     * @param fs state identifier
     * @return true if the state was not allready stored
     */
    public synchronized boolean insertRoot(FlyState fs) {
        return insert(fs, null);
    }

    /**
     * Remember a new state. The state has a parent in this area. The parent data
     * is remembered by the state itself, but its member variable is set by this
     * method. 
     * 
     * @param fs state identifier
     * @param old parent state identifier
     * @return true if the state was not allready stored.
     */
    public synchronized boolean insert(FlyState fs, FlyState old) {
        
        if (visited.containsKey(fs)) {
            return false;
        }
        
        visited.put(fs, fs);
        if (old != null) {
            fs.parentData = old.data;
        } else {
            // keep the old value. This is null for the initial state or
            // valid data for the state coming from another machine
        }
        
        if (visited.size() % sizeNotification == 0) {
            System.out.println(message + visited.size());
        }
        
        return true;
    }
    
    /**
     * Create list of state identifiers. The identifiers defines a path from the
     * given state to the one of the roots. If the last state identifier ( root
     * identifier ) contains null in its parentData member it is an initial state.
     * If the parentData member contains the same value as data member the callstack
     * continues in another area. 
     * @param fs state identifier - end of the callstack
     * @return list of state identifiers
     */
    public List<FlyState> getStackTrace(FlyState fs) {
        LinkedList<FlyState> ret = new LinkedList<FlyState>();
        
        FlyState parentDataKeeper = new FlyState();

        parentDataKeeper.data = fs.data;
        
        while (parentDataKeeper.data != null) {
            fs = visited.get(parentDataKeeper);
            if (fs == null) {
                parentDataKeeper.parentData = parentDataKeeper.data;
                ret.addFirst(parentDataKeeper);
                return ret;
            }
            ret.addFirst(fs);
            parentDataKeeper.data = fs.parentData;
        }
        
        return ret;
    }
    
    /**
     * Return true if given state was already visited
     * @param fs state
     * @return true if the state is member of this structure
     */
    public synchronized boolean contains(FlyState fs) {
        
        return visited.containsKey(fs);
    }
    
    
    public synchronized Set<FlyState> getVisitedStates(){
        return visited.keySet();
    }
    
    
    protected String message;

    public int getSizeNotification() {
        return sizeNotification;
    }
    
    
    private HashMap<FlyState, FlyState> visited;
    
}
