/* $Id$
 *
 * dChecker - Distributed Behavior Protocols Checker 
 * Copyright (C) 2006-2007  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.ow2.dsrg.fm.dchecker.distributed;


import org.ow2.dsrg.fm.dchecker.fly.FlyFSM;
import org.ow2.dsrg.fm.dchecker.fly.HierarchicalFlyFSM;
import org.ow2.dsrg.fm.dchecker.fly.FlyState;


/**
 * This class remembers upper bound of all visited states for each part of 
 * their identifier.
 * @author thorm
 */
public class VisitedNodesWithUpperBound extends VisitedNodes {
    
    /**
     * Creates a new instance of VisitedNodes
     * @param dim Number of parts the identifier consists of. This is the number of protocols in 
     * architecture descritption
     */
    public VisitedNodesWithUpperBound(String m,int dim) {
        super(m);
        visitedUpperBound = new int[dim];
    }
    
    public synchronized boolean insertRoot(FlyState fs, FlyFSM ffsm) {
        int[] splitted = new int[ffsm.getDimension()];
        ffsm.getStateParts(fs,splitted);
        return insertRoot(fs, splitted);
    }
    
    public synchronized boolean insertRoot(FlyState fs, int[] splitted) {
        return insert(fs, null, splitted);
    }
    
    /**
     * Insert the new state into the structure and change upper bounds if needed
     * @param fs new state identifier
     * @param splitted identifier splitted into parts
     */
    public synchronized boolean insert(FlyState fs, FlyState old, int[] splitted) {
        boolean ret = super.insert(fs, old);

        if (ret) {
            updateBound(fs, splitted);
        }
        return ret;
    }
    
    private void updateBound(FlyState fs, int[] splitted) {
        for (int i = 0; i < splitted.length; i++) {
            visitedUpperBound[i] = (visitedUpperBound[i] > splitted[i])
                    ? visitedUpperBound[i]
                    : splitted[i];
        }
        
    }
    
    public synchronized int getBound(int axis){
        return visitedUpperBound[axis];
    }
    
    
    /**
     * Insert the new state into the structure and change upper bounds if needed
     * @param fs new state identifier
     * @param splitted identifier splitted into parts
     */
    public synchronized void insert(FlyState fs, FlyState old, HierarchicalFlyFSM fsm) {
        int[] splitted = new int[fsm.getDimension()];

        insert(fs, old, splitted);
    }    

    /**
     * Returns true if the i-th part of given splitted state is bigger than i-th upper
     * bound
     * @param direction index of bound to be compared
     * @param splitted state identifier splitted into parts
     * @return true if the given identifier is out of the bounded area in particular direction
     */
    public synchronized boolean aboveBound(int direction, int[] splitted) {
        return visitedUpperBound[direction] < splitted[direction];
    }
    
    /**
     * Return true if the given splitted identifier is out of upper bounds
     * @param splitted state identifier splitted into parts corresponding to in-memory automatons states
     * @return true if the given state identifier is out of the bounds
     */
    public synchronized int aboveBound(int[] splitted) {
        for (int i = 0; i < splitted.length; i++) {
            if (visitedUpperBound[i] < splitted[i]) {
                return i;
            }
        }
        return -1;
    }
    
    public synchronized int howMuchAbove(int direction, int[] splitted) {
        return splitted[direction] - visitedUpperBound[direction];
    }
    
    /**
     * Return index of smallest bound
     * @return index of smallest bound
     */
    public synchronized int thinnestDirection() {
        int min = visitedUpperBound[0];
        int ret = 0;

        for (int i = 0; i < visitedUpperBound.length; i++) {
            if (min > visitedUpperBound[i]) {
                min = visitedUpperBound[i];
                ret = i;
            }
        }
        return ret;
    }
    
    /**
     * i-th value is upper bound of all i-th parts of state identifiers stored in this
     * object
     */
    private int[] visitedUpperBound;
}
