/* $Id$
 *
 * dChecker - Distributed Behavior Protocols Checker 
 * Copyright (C) 2006-2007  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.ow2.dsrg.fm.dchecker.distributed.spacemap;


import java.io.Serializable;


/**
 * Bounding box of n-dimensional space
 * @author Tomas Poch
 */
public class BoundingBox implements Serializable {
    
    /**
     * Creates a new instance of SpaceInterval
     * @param min coordinates of first corner
     * @param max coordinates of second corner
     */
    public BoundingBox(int[] min, int[] max) {
        minimal = min;
        maximal = max;
    }
    
    /**
     * Creates a new instance of SpaceInterval
     * @param dimension Dimension of state space.
     * The bounding box is set to [0,0,...,0] - [0,0,...,0]
     */
    public BoundingBox(int dimension) {
        minimal = new int[dimension];
        maximal = new int[dimension];
    }
    
    /**
     * Obtain the member of bounding box that is closest to the begining
     * @return coordinates of the first corner
     */
    public int[] getMinimalCoord() {
        return minimal;
    }

    /**
     * Obtain the member of bounding box such that all other points from the bounding box
     * are closer to the begining
     * @return coordinates of the second corner
     */
    public int[] getMaximalCoord() {
        return maximal;
    }
    
    /**
     * Split this box into n parts along the given axis.
     * The original bounding box should be forgotten after it.
     * @param axis axis that is going to be divided
     * @param parts number of new parts.
     * @return newly created boxes
     */
    public BoundingBox[] split(int axis, int parts) {
        int base = minimal[axis];
        int step = (maximal[axis] - base) / parts;
        // Null protocol
        if (step==0){
            BoundingBox[] ret = new BoundingBox[1];
            ret[0] = new BoundingBox(minimal.clone(),maximal.clone());
            return ret;
        }
        
        BoundingBox[] ret = new BoundingBox[parts];

        
        for (int i = 0; i < parts - 1; i++) {
            int[] newMin = minimal.clone();
            int[] newMax = maximal.clone();

            newMin[axis] = base + (i * step);
            newMax[axis] = base + ((i + 1) * step);
            ret[i] = new BoundingBox(newMin, newMax);
        }
        
        int[] newMin = minimal.clone();

        newMin[axis] = base + ((parts - 1) * step);
        ret[(parts - 1)] = new BoundingBox(newMin, maximal);
        return ret;
    }
    
    /**
     * Split the box into two parts along given axis
     * This object than represents the part which is closer to zero.
     * @param axis axis used to split the box
     * @param where splitting point on the axis
     * @return new box
     */
    public BoundingBox splitAt(int axis, int where) {
        int[] newMin = minimal.clone();

        newMin[axis] = where;
        int[] newMax = maximal.clone();

        maximal[axis] = where;
        return new BoundingBox(newMin, newMax);
    }
    
    /**
     * Get index of the axis which correspond to the widest edge of this box
     * @return index of the axis
     */
    public int getWidestAxis() {
        int max = 0;
        int ret = 0;
        int d;

        for (int i = 0; i < maximal.length; i++) {
            d = maximal[i] - minimal[i];
            if (max <= d) {
                ret = i;
                max = d;
            }
        }
        return ret;
    }
    
    /**
     * Get index of the axis which correspond to the most narrow edge of this box
     * @return index of the axis
     */
    public int getNarrowestAxis() {
        int min = Integer.MAX_VALUE;
        int ret = 0;
        int d;

        for (int i = 0; i < maximal.length; i++) {
            d = maximal[i] - minimal[i];
            if (min >= maximal[i]) {
                ret = i;
                min = maximal[i];
            }
        }
        return ret;
    }
    
    /**
     * Get width of the box along the given axis
     * @param axis index of the axis
     * @return width along the given axis
     */
    public int getWidth(int axis) {
        return maximal[axis] - minimal[axis];
    }
    
    /**
     * Get volume of bounded box
     * @return volume
     */
    public long getVolume() {
        long ret = 1;

        for (int i = 0; i < minimal.length; i++) {
            ret *= maximal[i] - minimal[i];
        }
        return ret;
    }
    
    /**
     * Return <code>true</code> if the given coordinates are member of the bounding box
     * @param splittedState state identifier splitted into the parts
     * @return <code>true</code> if the given coordinates are member of the bounding box
     */
    public boolean belongs(int[] splittedState) {
        for (int i = 0; i < minimal.length; i++) {
            int ithPart = splittedState[i];

            if (!((minimal[i] <= ithPart) && (ithPart < maximal[i]))) {
                return false;
            }
        }
        return true;
    }
    
    /**
     * coordinates of the first corner
     */
    protected int[] minimal;

    /**
     * coordinates of the second corner
     */
    protected int[] maximal;
    
    
    public String toString(){
        StringBuffer ret = new StringBuffer();
        ret.append("[");
        for (int i = 0; i<maximal.length;i++){
            ret.append(minimal[i]);
            ret.append("-");
            ret.append(maximal[i]);
            if (i!=maximal.length-1){
                ret.append(",");
            }
        }
        
        ret.append("]");
        return ret.toString();
    
    }
}
