/* $Id$
 *
 * dChecker - Distributed Behavior Protocols Checker 
 * Copyright (C) 2006-2007  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.ow2.dsrg.fm.dchecker.distributed.spacemap;



import java.util.ArrayList;
import java.util.LinkedList;
import java.util.TreeMap;
import org.ow2.dsrg.fm.bptools.util.Pair;


/**
 *
 * Represents the function from state space of automaton to T
 * The state space is divided into areas that has assigned come value of type T
 *
 * This case implements one dimensional case of SpaceMap
 */
 
 
public class LeafSpaceMap<T> extends SpaceMapImpl<T> {
    
    /**
     * Creates a new instance of DistributionInfo
     */
    public LeafSpaceMap() {
        tmpRange = new Range(0, 0);
        map = new TreeMap<Range<T>, Range<T>>();
    }
    
    /**
     *
     * Assign value to the part of state space
     * @param keyRange range the value is assigned to. Only the first dimension 
     * of this bounding box is considered as this is class implements one-dimensional
     * case
     * @param value assigned value
     *
     */
    public void addInterval(BoundingBox keyRange, T value) {
        Range<T> newRng = new Range<T>(keyRange.minimal[0], keyRange.maximal[0]);

        newRng.value = value;
        
        LinkedList<Pair<Integer, Integer> > done = new LinkedList<Pair<Integer, Integer> >();
        
        TreeMap<Range<T>, Range<T>> newMap = new TreeMap<Range<T>, Range<T>>();
        
        for (Range<T> old : map.keySet()) {
            if (newRng.noIntersection(old)) {
                newMap.put(old, old);
                continue;
            }
            
            if (newRng.isParameterSubrange(old)) {
                old.value = value;
                newMap.put(old, old);
                done.add(new Pair<Integer, Integer>(old.left, old.right));
                continue;
            }
            
            if (old.left < newRng.left) {
                Range<T> oldLeftPart = new Range<T>(old.left, newRng.left);

                oldLeftPart.value = old.value;
                newMap.put(oldLeftPart, oldLeftPart);
            }
            
            if (old.right > newRng.right) {
                Range<T> oldRightPart = new Range<T>(newRng.right, old.right);

                oldRightPart.value = old.value;
                newMap.put(oldRightPart, oldRightPart);
            }
            
            if (newRng.left <= old.left) {
                Range<T> part = new Range<T>(old.left, newRng.right);

                part.value = value;
                
                newMap.put(part, part);
                done.add(new Pair<Integer, Integer>(part.left, part.right));
            }
            
            if (newRng.right >= old.right) {
                Range<T> part = new Range<T>(newRng.left, old.right);

                part.value = value;
                newMap.put(part, part);
                done.add(new Pair<Integer, Integer>(part.left, part.right));
            }
            
        }
        
        map = newMap;
        
        int gapLeft = newRng.left;
        int gapRight = 0;
        Pair<Integer, Integer> interval;

        while ((interval = done.poll()) != null) {
            gapRight = interval.first;
            if (gapLeft < gapRight) {
                fillGap(gapLeft, gapRight, value);
            }
            gapLeft = interval.second;
        }
        gapRight = newRng.right;
        if (gapLeft < gapRight) {
            fillGap(gapLeft, gapRight, value);
        }
        

    }
    
    /**
     * Create a new record.
     * @param left start of the record
     * @param right end of the record
     * @param value value assigned to this part of space
     */
    private void fillGap(int left, int right, T value) {
        Range<T> part = new Range<T>(left, right);

        part.value = value;
        map.put(part, part);
    }
    
    /**
     * Get value assigned to the given coordinates
     * As this is one-dimensional case the key is treated as one dimensional.
     * @param key coordinates - only the first value is considered
     * @return assigned value
     */
    public T get(int[] key) {
        tmpRange.update(key[0]);
        Range<T> r = map.get(tmpRange);
        
        return (r == null) ? null : r.value;
        
    }
    
    public boolean changeValueForWholeArea(int[] key,T value){
        tmpRange.update(key[0]);
        Range<T> r = map.get(tmpRange);
        if (r!=null) {
            r.value =value;
            return true;
        }
        return false;
    }
    
    
    /**
     * Create deep copy of structure.
     * The result has the same references to assigned value as source.
     * @return new instance containing the same intervals
     */
    public SpaceMapImpl<T> deepCopy() {
        LeafSpaceMap<T> ret = new LeafSpaceMap<T>();

        for (Range<T> i:map.keySet()) {
            Range<T> rngCopy = new Range<T>(i.left, i.right);

            rngCopy.value = i.value;
            ret.map.put(rngCopy, rngCopy);
        }
        return ret;
    }
    
    /**
     * Used to print out the content of this structure.
     *
     * For each item of the map concat the prefixes, append the value and store
     * the result into lines.
     *
     * @param lines lines keeping the records that were already printed out
     * @param prefixes actual list of prefixes - ranges in higher dimensions
     */
    public void toStringList(ArrayList<String> lines, LinkedList<String> prefixes) {
        
        for (Range<T>mapItem :map.keySet()) {
            StringBuffer sb = new StringBuffer("[");

            sb.append(mapItem.left);
            sb.append("-");
            sb.append(mapItem.right);
            
            for (String s:prefixes) {
                sb.append("\t,");
                sb.append(s);
            }
            sb.append("]:");
            if (mapItem.value != null) {
                sb.append(mapItem.value.toString());
            } else {
                sb.append("null");
            }
            
            lines.add(sb.toString());
        }
        
    }
    
    /**
     * Contains the very data
     */
    TreeMap<Range<T>, Range<T>> map;
    
    /**
     * Instead of local variables to avoid garbage collection
     */
    Range<T> tmpRange;
    
}
