/* $Id$
 *
 * dChecker - Distributed Behavior Protocols Checker 
 * Copyright (C) 2006-2007  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.ow2.dsrg.fm.dchecker.distributed.spacemap;



import java.util.ArrayList;
import java.util.LinkedList;
import java.util.TreeMap;

import org.ow2.dsrg.fm.bptools.util.Pair;


/**
 * Represents the function from state space of cartesian product of automatons to
 * T
 * The state space is divided into areas that has assigned come value of type T
 *
 * The number of automatons can be from 2 to n
 *
 *
 * The function is implemented as function from interval to function from state
 * space of dimension n-1 to T
 * @author Tomas Poch
 */
public class NDimSpaceMap<T> extends SpaceMapImpl<T> {
    
    /**
     * Creates a new instance of NDimSpaceMap
     * @param dimension dimension of state space<br>
     * Number of protocols in the architecture
     */
    public NDimSpaceMap(int dimension) {
        assert dimension >= 2;
        this.dimension = dimension;
        tmpRange = new Range<SpaceMap<T>>(0, 0);
        map = new TreeMap<Range<SpaceMapImpl<T>>, Range<SpaceMapImpl<T>>>();
    }
    
    /**
     * Assign value to the part of state space
     * @param keyRange range the value is assigned to.
     * @param value assigned value
     */
    public void addInterval(BoundingBox keyRange, T value) {
        Range<SpaceMapImpl<T>> newRng = new Range<SpaceMapImpl<T>>(
                keyRange.minimal[dimension - 1], keyRange.maximal[dimension - 1]);
        
        LinkedList<Pair<Integer, Integer> > done = new LinkedList<Pair<Integer, Integer> >();
        
        TreeMap<Range<SpaceMapImpl<T>>, Range<SpaceMapImpl<T>>> newMap = new TreeMap<Range<SpaceMapImpl<T>>, Range<SpaceMapImpl<T>>>();
        
        for (Range<SpaceMapImpl<T>> old : map.keySet()) {
            if (newRng.noIntersection(old)) {
                newMap.put(old, old);
                continue;
            }
            
            if (newRng.isParameterSubrange(old)) {
                old.value.addInterval(keyRange, value);
                newMap.put(old, old);
                done.add(new Pair<Integer, Integer>(old.left, old.right));
                continue;
            }
            
            if (old.left < newRng.left) {
                Range<SpaceMapImpl<T>> oldLeftPart = new Range<SpaceMapImpl<T>>(
                        old.left, newRng.left);

                oldLeftPart.value = old.value.deepCopy();
                newMap.put(oldLeftPart, oldLeftPart);
            }
            
            if (old.right > newRng.right) {
                Range<SpaceMapImpl<T>> oldRightPart = new Range<SpaceMapImpl<T>>(
                        newRng.right, old.right);

                oldRightPart.value = old.value.deepCopy();
                newMap.put(oldRightPart, oldRightPart);
            }
            
            if (newRng.left <= old.left) {
                Range<SpaceMapImpl<T>> part = new Range<SpaceMapImpl<T>>(
                        old.left, newRng.right);

                part.value = old.value.deepCopy();
                part.value.addInterval(keyRange, value);
                newMap.put(part, part);
                done.add(new Pair<Integer, Integer>(part.left, part.right));
            }
            
            if (newRng.right >= old.right) {
                Range<SpaceMapImpl<T>> part = new Range<SpaceMapImpl<T>>(
                        newRng.left, old.right);

                part.value = old.value.deepCopy();
                part.value.addInterval(keyRange, value);
                newMap.put(part, part);
                done.add(new Pair<Integer, Integer>(part.left, part.right));
            }
            
        }

        map = newMap;
        
        int gapLeft = newRng.left;
        int gapRight = 0;
        Pair<Integer, Integer> interval;

        while ((interval = done.poll()) != null) {
            gapRight = interval.first;
            if (gapLeft < gapRight) {
                fillGap(gapLeft, gapRight, value, keyRange);
            }
            gapLeft = interval.second;
        }
        gapRight = newRng.right;
        if (gapLeft < gapRight) {
            fillGap(gapLeft, gapRight, value, keyRange);
        }
        

    }
    
    /**
     * Create a new record.
     * @param left start of the record
     * @param right end of the record
     * @param value value assigned to this part of space
     * @param keyRange information about the lower dimensions
     */
    private void fillGap(int left, int right, T value, BoundingBox keyRange) {
        Range<SpaceMapImpl<T>> part = new Range<SpaceMapImpl<T>>(left, right);

        if (dimension > 2) {
            part.value = new NDimSpaceMap<T>(dimension - 1);
        } else {
            part.value = new LeafSpaceMap<T>();
        }
        part.value.addInterval(keyRange, value);
        map.put(part, part);
    }
    
    /**
     * Get value assigned to the given coordinates
     * @param key coordinates
     * @return assigned value
     */
    public T get(int[] key) {
        tmpRange.update(key[dimension - 1]);
        Range<SpaceMapImpl<T> > r = map.get(tmpRange);
        
        return (r == null) ? null : r.value.get(key);
    }
    
    public boolean changeValueForWholeArea(int[] key,T value){
        tmpRange.update(key[dimension - 1]);
        Range<SpaceMapImpl<T> > r = map.get(tmpRange);
        if(r != null){
            return r.value.changeValueForWholeArea(key,value);
        }
        return false;
    }
    
    /**
     * Create deep copy of structure.
     * The assigned values from T are not copied deeply
     * @return new instance containing areas of the same sizes and positions
     */
    public SpaceMapImpl<T> deepCopy() {
        NDimSpaceMap<T> ret = new NDimSpaceMap<T>(dimension);
        
        for (Range<SpaceMapImpl<T>> i:map.keySet()) {
            Range<SpaceMapImpl<T>> rngCopy = new Range<SpaceMapImpl<T>>(i.left,
                    i.right);
            
            rngCopy.value = i.value.deepCopy();
            ret.map.put(rngCopy, rngCopy);
        }
        return ret;
    }
    
    /**
     * Used to print out the content of this structure.
     *
     * For each item of the map append the range to the prefixes and call recursively
     * this method on the value of the range.
     *
     * @param lines lines keeping the records that were already printed out
     * @param prefixes actual list of prefixes - ranges in higher dimensions
     */
    public void toStringList(ArrayList<String> lines, LinkedList<String> prefixes) {
        
        for (Range<SpaceMapImpl<T> >mapItem :map.keySet()) {
            prefixes.addFirst(mapItem.left + "-" + mapItem.right);
            mapItem.value.toStringList(lines, prefixes);
            prefixes.removeFirst();
        }
        
    }
    
    /**
     * To decrease load of garbage collector
     */
    Range<SpaceMap<T>> tmpRange;
    
    /**
     * Function from interval to function from state space of smaller dimension
     */
    TreeMap<Range<SpaceMapImpl<T>>, Range<SpaceMapImpl<T>>> map;
    
    /**
     * Dimension of state space
     */
    int dimension;
}
