/* $Id$
 *
 * dChecker - Distributed Behavior Protocols Checker 
 * Copyright (C) 2006-2007  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.ow2.dsrg.fm.dchecker.fly;


import java.io.PrintStream;
import java.util.Set;
import org.ow2.dsrg.fm.bptools.BPEventTableImpl;
import org.ow2.dsrg.fm.bptools.IEventTable;
import org.ow2.dsrg.fm.bptools.fsm.State;
import org.ow2.dsrg.fm.dchecker.distributed.spacemap.BoundingBox;
import org.ow2.dsrg.fm.dchecker.errors.BadActivityException;
import org.ow2.dsrg.fm.dchecker.errors.CompositionException;
import org.ow2.dsrg.fm.dchecker.errors.NondeterministicConsentException;


/**
 * Implementation of consent operator.
 * Takes two FlyFSMs and generate composed state space on the fly
 *
 * The particular state is identified by the concated identifier of both childrens
 * encoded into long
 *
 * The structure of whole architecture looks like this:
 * <pre>
 *          Consent(bindings)
 *           /          \
 *          /            \
 *     SimpleFSM   Consent(bindings)
 *                  /         \
 *                 /           \
 *           SimpleFSM   Consent(bindings)
 *                          /         \
 *                              .....
 *                        /             \
 *                    SimpleFSM       SimpleFSM
 *
 * </pre>
 * @author thorm
 */
public class FlyBinaryConsent extends HierarchicalFlyFSM {
    
    /**
     * Create new instance of FlyConsent
     * @param l first automaton ( in memory explicit automaton )
     * @param r second automaton (other consent operator or )
     * @param bindings bindings between two automatons passed to the function
     */
    public FlyBinaryConsent(SimpleFlyFSM l, HierarchicalFlyFSM r, Set<Integer> bindings, Set<Integer> unbound) {
        super(unbound);
        this.l = l;
        this.r = r;
        this.bindings = bindings;
        dimension = r.dimension + 1;
    }
    
    /**
     * Initialize member variables used to encode state identifier into long number
     *
     * This invokes the same method on its childs.
     * @param base number that is used to divide long identifier to obtain identifier
     *        of state generated by this consent
     */
    public void setBase(BitPartInfo base) {
        l.partIndex = partIndex;
        r.partIndex = partIndex + 1;
        
        l.setBase(base);
        r.setBase(base.multiply(l.getMaxState()));
    }
    
    /**
     * Set the initial state.
     * @param state store the initial state here.
     */
    public void init(FlyState state) {
        l.init(state);
        r.init(state);
    }
    
    
    public int identifierSize(){
        return r.identifierSize();
    }
    
    
    /**
     * Get the next symbol that labels some edge coming from given state
     * @param state state identifier
     * @param last last symbol obtained by this method
     * @return smallest symbol that is bigger than last
     */
    public int getNextSymbol(FlyState state, int last) throws CompositionException {
        int lnext = last, rnext = last;
        int min = last;
        int methodCode;
        do {
            if (min == lnext) { 
                lnext = l.getNextSymbol(state, min);
            }
            if (min == rnext) { 
                rnext = r.getNextSymbol(state, min);
            }
            
            if (lnext != State.AFTER_LAST_EDGE && lnext == rnext) {
                throw new NondeterministicConsentException(state, lnext);
            }
            
            min = (lnext < rnext) ? lnext : rnext;
            
        } while (min != State.AFTER_LAST_EDGE && BPEventTableImpl.staticIsAccept(min)
                && (bindings.contains(min)||unbound.contains(min))
                );
        
        return min;
        
    }
    
    /**
     * Fill the newState with identifier of state that can be reached from oldState using
     * an edge labeled by symbol.
     *
     * @param oldState identifier of edge source state
     * @param newState initially must contain the same identifier as oldState which is modified
     *                 to the identifier of the target state.
     * @param symbol edge label
     * @throws bpchecker.errors.CompositionException if composition error occure
     * @return true if the transition is valid
     */
    public boolean follow(FlyState oldState, FlyState newState, int symbol) throws CompositionException {
        
        if (bindings.contains(symbol)) {
            
            int complement = BPEventTableImpl.getComplement(symbol);
            
            if (l.follow(oldState, newState, symbol)) {
                if (r.follow(oldState, newState, complement)) {
                    return true;
                } else {
                    throw new BadActivityException(oldState, symbol);
                }
            } else if (r.follow(oldState, newState, symbol)) {
                if (l.follow(oldState, newState, complement)) {
                    return true;
                } else {
                    throw new BadActivityException(oldState, symbol);
                }
            }
            
            assert false;
            
        }
        
        if (l.follow(oldState, newState, symbol)) {
            return true;
        }
        return r.follow(oldState, newState, symbol);
        
    }
    
    /**
     * Return true if the given state is final.
     * @param s state
     * @return true iff the given state is final
     */
    public boolean isFinal(FlyState s) {
        return l.isFinal(s) && r.isFinal(s);
    }
    
    /**
     * Fill the given array with identifiers of actual states of the automatons
     * the consent operator consist from
     * Invokes this method on the children
     *
     * @param state state identifier that consist of identifiers of all automatons
     * @param parts array of the correct size used to store the result
     */
    public void getStateParts(FlyState state, int[] parts) {
        l.getStateParts(state, parts);
        r.getStateParts(state, parts);
    }
    
    /**
     * Set the size of the given bounding box to contain whole state space
     * Invokes this method on the children
     *
     * @param bb to be filled
     */
    public void fillBoundingBox(BoundingBox bb) {
        l.fillBoundingBox(bb);
        r.fillBoundingBox(bb);
    }
    
    public int[][] getEdgeDensity(){
        int[][] ret;
        if (partIndex == dimension-1){
            ret = new int[dimension][];
        }else{
            ret = r.getEdgeDensity();
        }
        ret[partIndex] = l.getEdgeDensity()[0];
        return ret;
    }
    
    /**
     * The left child of the consent operator
     */
    SimpleFlyFSM l;
    
    /**
     * The right child of the consent operator
     */
    HierarchicalFlyFSM r;
    
    /**
     * Bindings of this consent operator
     */
    Set<Integer> bindings;
    
    public void dump(PrintStream os, IEventTable et){
            l.dump(os, et);
            os.println("sync{");
            dumpEventList(os,et,bindings);
            os.println("}");
            r.dump(os, et);
        }
    
}
