/* $Id$
 *
 * dChecker - Distributed Behavior Protocols Checker 
 * Copyright (C) 2006-2007  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.ow2.dsrg.fm.dchecker.fly;



import java.util.Set;
import org.ow2.dsrg.fm.bptools.fsm.State;
import org.ow2.dsrg.fm.dchecker.errors.BadActivityException;
import org.ow2.dsrg.fm.dchecker.errors.CompositionException;
import org.ow2.dsrg.fm.dchecker.errors.NoActivityException;


/**
 * The descendants can generate the state space on the fly
 * @author thorm
 */

public abstract class HierarchicalFlyFSM extends FlyFSM {
    public HierarchicalFlyFSM(Set<Integer> unbound) {
        super(unbound);
    }
    
    /**
     * Initialize member variables used to encode state identifier into long number
     * @param base number that is used to divide long identifier to obtain identifier
     *        of state generated by this consent
     */
    abstract public void setBase(BitPartInfo base);
    
    /**
     * Get the next symbol that labels some edge coming from given state
     * @param state state identifier
     * @param last last symbol obtained by this method
     * @return smallest symbol that is bigger than last
     */
    abstract public int getNextSymbol(FlyState state, int last) throws CompositionException;
    
    /**
     * Fill the newState with identifier of state that can be reached from oldState using
     * an edge labeled by symbol.
     *
     * @param oldState identifier of edge source state
     * @param newState initially must contain the same identifier as oldState which is modified 
     *                 to the identifier of the target state.
     * @param symbol edge label
     * @throws bpchecker.errors.CompositionException if composition error occure
     * @return true if the transition is valid
     */
    abstract public boolean follow(FlyState oldState, FlyState newState, int symbol) throws CompositionException;

    /**
     * Fill the given list with identifiers of actual states of the automatons
     * the consent operator consist from
     * @param f state identifier that consist of identifiers of all automatons
     * @param l list to store result
     */
    // abstract public void getStateParts(FlyState f,List<Integer> l);
    
    
    /**
     * Fill the given array with identifiers of actual states of the automatons
     * the consent operator consist from
     *
     * @param state state identifier that consist of identifiers of all automatons
     * @param parts array of the correct size used to store the result
     */
    public abstract void getStateParts(FlyState state, int[] parts);
    
    /**
     * index of the automaton in the architecture
     */
    protected int partIndex = 0;

    /**
     * number that is used to divide long identifier to obtain identifier
     * of state generated by this automaton
     */
    protected BitPartInfo partInfo;

    public StateNeighbourEnumeration getNeighboursEnum(FlyState fs) {
        return new StateNeighbourEnumeration(fs) {

            public FlyState nextElement() throws CompositionException {
                FlyState ret = new FlyState(f);
                boolean followOk = follow(f, ret, last);

                assert followOk;
                return ret;
            }

            public boolean hasMoreElements() throws CompositionException {
                int next = getNextSymbol(f, last);

                if (next == State.AFTER_LAST_EDGE
                        && last == State.BEFORE_FIRST_EDGE && !isFinal(f)) {
                    throw new NoActivityException(f);
                }
                // TODO: check this : unbound is list of events, no methods so I removed the translation
                if (unbound.contains(next)) {
                    throw new BadActivityException(f, next);
                }
                last = next;
                return last != State.AFTER_LAST_EDGE;
            }

            int last = State.BEFORE_FIRST_EDGE;
        };
    }

    /**
     * Return the number of automatons this automaton consist from
     * 
     * @return number of automatons
     */
    public int getDimension() {
        return dimension;
    }

    /**
     * number of automatons this automaton consist from
     */
    protected int dimension;

}
