/* $Id$
 *
 * dChecker - Distributed Behavior Protocols Checker 
 * Copyright (C) 2006-2007  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.ow2.dsrg.fm.dchecker.fly;


import java.io.PrintStream;
import java.util.Set;
import org.ow2.dsrg.fm.bptools.IEventTable;
import org.ow2.dsrg.fm.bptools.fsm.CompactFSM;
import org.ow2.dsrg.fm.bptools.fsm.SimpleState;
import org.ow2.dsrg.fm.bptools.fsm.State;
import org.ow2.dsrg.fm.dchecker.distributed.spacemap.BoundingBox;


/**
 * Explicit representation of FlyFSM representing one protocol. Such automatons 
 * are connected by ConsentFSM
 * @author thorm
 */
public class SimpleFlyFSM extends HierarchicalFlyFSM {
    
    /**
     * Creates a new instance of IdentifiersFSM
     * @param start identifier of start state
     * @param table table of states
     */
    public SimpleFlyFSM(int start, SimpleState[] table, Set<Integer> unbound) {
        super(unbound);
        this.table = table;
        this.start = start;
        dimension = 1;
        
        maxState = table.length;
        table_base = 0;
        whole = true;
    }
    
    public SimpleFlyFSM(int _base, int _maxState, int start, SimpleState[] table, Set<Integer> unbound) {
        this(start, table, unbound);
        table_base = _base;
        whole = false;
        maxState = _maxState;
    }
    
    public SimpleFlyFSM(CompactFSM cfsm, Set<Integer> unbound){
        this(cfsm.getStartState(), cfsm.getStateTable(), unbound);
    }
    
    /**
     * Initialize member variables used to encode state identifier into long number
     * @param base number that is used to divide long identifier to obtain identifier
     *        of state generated by this consent
     *        The result is moduled by number of states and it is the identifier
     */
    public void setBase(BitPartInfo base) {
        
        this.partInfo = base;
        base.setUpperBound(maxState);
        
    }
    
    /**
     * Modify the given identier so that the part that is assigned to this automaton
     * contains the initial state
     * @param state state identifier of whole architecture
     */
    public void init(FlyState state) {
        state.resize(partInfo.firstBitPosition + partInfo.width);
        setStatePart(state, start);
    }
    
    public int identifierSize(){
        return partInfo.firstBitPosition + partInfo.width;
    }
    
    /**
     * Get the next symbol that labels some edge coming from given state
     * @param state state identifier
     * @param last last symbol obtained by this method
     * @return smallest symbol that is bigger than last
     */
    public int getNextSymbol(FlyState state, int last) {
        int localStateId = getStatePart(state);
        int index = localStateId - table_base;
        assert (index<table.length): "Index: "+ index+
                                    " Table length: "+ table.length + 
                                    " Table base: " + table_base + 
                                    " localStateId: " +localStateId +
                                    " partInfo: ["+ partInfo.firstBitPosition +","+ Long.toBinaryString(partInfo.mask) +","+ partInfo.width + "]";
        SimpleState s = table[index];
        int ret = s.nextEdge(last);

        return ret;
    }
    
    public boolean accepts(FlyState oldState, int symbol) {
        int localState = getStatePart(oldState);
        int nextState = table[localState - table_base].getEdgeTarget(symbol);

        return (nextState >= 0);
    }
    
    /**
     * Fill the newState with identifier of state that can be reached from oldState using
     * an edge labeled by symbol.
     * 
     * @return true if the transition is valid
     * @param oldState identifier of edge source state
     * @param newState initially must contain the same identifier as oldState which is modified 
     *                 to the identifier of the target state.
     * @param symbol edge label
     */
    public boolean follow(FlyState oldState, FlyState newState, int symbol) {
        int localState = getStatePart(oldState);
        int nextState = table[localState - table_base].getEdgeTarget(symbol);

        if (nextState < 0) {
            return false;
        }
        setStatePart(newState, nextState);
        return true;
    }
    
    /**
     * Get the part of the identifier that identifies actual state of this automaton
     * @param state identifier of whole architecture state
     * @return actual state of this automaton
     */
    public int getStatePart(FlyState state) {
        return state.getPart(partInfo);
        // return state.id.divide(base).mod(maxState).intValue();
    }
    
    /**
     * Set the part of the identifier that identifies actual state of this automaton
     * @param part identifier of state of this machine 
     * @param state identifier of whole architecture state
     */
    public void setStatePart(FlyState state, int part) {
        
        /* BigInt lowerPart = state.mod(base);
         BigInt higherPart = state.divide(baseTimesmaxState);*/
        state.setPart(part, partInfo);
        // state.data = higherPart.multiplyLocal(maxState).addLocal(bigPart).multiply(base).add(lowerPart);
        // assert state.data.compareTo(BigInteger.ZERO)>=0;
    }
    
    /**
     * Return true if the part of the given state that corresponds to this 
     * automaton is one of its final states.
     * @param s architecture state
     * @return true if this automaton is in final state
     */
    public boolean isFinal(FlyState s) {
        int localState = getStatePart(s);

        return table[localState - table_base].isFinal();
    }
    
    /**
     * Fill the given array with identifiers of actual states of the automatons
     * the consent operator consist from
     *
     * @param state state identifier that consist of identifiers of all automatons
     * @param parts array of the correct size used to store the result
     */
    public void getStateParts(FlyState state, int[] parts) {
        parts[partIndex] = getStatePart(state);
    }
    
    /**
     * Set the size of the given bounding box to contain whole state space
     * @param bb to be filled
     */
    public void fillBoundingBox(BoundingBox bb) {
        
        if (whole){
            bb.getMinimalCoord()[partIndex] = 0;
            bb.getMaximalCoord()[partIndex] = table.length;
        }else {
            bb.getMinimalCoord()[partIndex] = table_base;
            bb.getMaximalCoord()[partIndex] = table.length+table_base;
        }
    }
    
    public int getEdgeLabelConnecting(int state1, int state2) {
        int ret = table[state1 - table_base].getLabelForTarget(state2);

        assert ret != State.NO_SUCH_EDGE:"Not existing edge: " + state1 + " -> "
                + state2;
        return  ret;
        
    }
    
    public SimpleFlyFSM getSlice(int min, int max) {
        assert whole;
        SimpleState[] newTable = new SimpleState[max - min];

        for (int i = 0; i < max - min; i++) {
            newTable[i] = table[min + i];
        }
        SimpleFlyFSM ret = new SimpleFlyFSM(min, maxState, start, newTable,
                unbound);

        return ret;
    }
       
    
    public int[][] getEdgeDensity(){
        int[] vect = new int[table.length];
        for(int s1id = 0 ; s1id<table.length; s1id++){
            SimpleState s1 = table[s1id];
            int edgeLabel = State.BEFORE_FIRST_EDGE;
            edgeLabel = s1.nextEdge(edgeLabel);
            while (edgeLabel!=State.AFTER_LAST_EDGE){
                int s2id = s1.getEdgeTarget(edgeLabel);
                if (s1id<s2id){
                    vect[s1id]++;
                    vect[s2id]--;
                }else{
                    vect[s2id]++;
                    vect[s1id]--;
                }
                edgeLabel = s1.nextEdge(edgeLabel);
            }
        }
        
        int sum = 0;
        for(int i = 0; i<vect.length; i++){
            sum+=vect[i];
            vect[i]=sum;
        }
        
        int[][] ret = new int[1][];
        ret[0] = vect;
        return ret;
    }
    
    /**
     * Table of states. Each State contains array of edges
     */
    SimpleState[] table;
    
    public int table_base;
    private boolean whole = true;
    
    /**
     * Initial state
     */
    int start;
    
    /**
     * Maximal local state identifier - number of states - 1
     */
    private int maxState;
    
    /**
     * Get maximal local state identifier
     * @return  maximal local state identifier
     */
    public int getMaxState() {
        return maxState;
    }
    
    
    public void dump(PrintStream os, IEventTable et){
        os.println("#initial: "+ start);
        for(int i=0;i<table.length;i++){
            
            os.println("State "+(i+table_base)+": [");
            SimpleState ss = table[i];
            int edgeLabel = State.BEFORE_FIRST_EDGE;
            while ((edgeLabel=ss.nextEdge(edgeLabel))!=State.AFTER_LAST_EDGE){
                String labelStr = eventString(et, edgeLabel);
                os.println("  --"+labelStr+"->"+ss.getEdgeTarget(edgeLabel)+",");
            }
            os.println("]");
        }
    }
}
