/* $Id$
 *
 * dChecker - Distributed Behavior Protocols Checker 
 * Copyright (C) 2006-2007  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.fm.dchecker.rmi;


import org.ow2.dsrg.fm.dchecker.distributed.Area;
import org.ow2.dsrg.fm.dchecker.distributed.FileDump;
import org.ow2.dsrg.fm.dchecker.errors.CompositionException;
import org.ow2.dsrg.fm.dchecker.fly.FlyState;
import java.io.IOException;
import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.Set;
import java.util.logging.Logger;
import org.ow2.dsrg.fm.bptools.Preferences;


/**
 *
 * @author thorm
 */
public abstract class RMIClientTaskImpl extends UnicastRemoteObject implements ClientTask {

    /** Creates a new instance of RMIClientTaskImpl */
    public RMIClientTaskImpl(RMIClientImpl client, ServerTask serverTask) throws RemoteException {
        super();
        this.serverTask = serverTask;
        this.client = client;
        cid = serverTask.getClientInitData();
        stateInputQueue = new StateInputQueue(cid, client.usedPort, client.clientId, this);
        stateOutputQueue = new StateOutputQueue(cid, client.usedPort, client.clientId);
    }

    public void init() throws IOException, ClientLostException {
        stateInputQueue.bind();
        stateOutputQueue.connectToOtherClients();
        stateInputQueue.acceptConnectionsFromOtherClients();
        stateInputQueue.start();
    }

    public synchronized void finalCleanup() throws NoSuchObjectException, IOException {
        while (!finishInvoked) {
            try {
                wait();
            } catch (InterruptedException ex) {
            }
        }
        while (!unexportObject(this, false)) {
        }
        stateInputQueue.close();
        stateOutputQueue.close();
    }

    /**
     * This method runs in the main thread and traverses the statespace.
     * If there is no work to do it notifies the server that this client is
     * iddle and waits until a new area or entry point emerges
     * from the server.
     *
     * The new areas are automaticly sent to iddle clients and new entry point
     * comes when some other client reaches the boundary of some area that belongs
     * to this client
     *
     * @throws java.lang.Exception
     */
    public void run() throws Exception {

        try {
            while (!end) {
                waitForServer();
                processNewAreas();
                processNewEntryPoints();

                boolean workDone = false;
                workDone = visitNStates(1000);
                if (!workDone) {
                    getIddle();
                }
            }

        } catch (CompositionException e) {
            serverTask.errorFound(e);
        }

    }

    public abstract boolean enqueNewEntryPoint(FlyState ep);
    public abstract void outOfArea(FlyState fs, int[] splitted) throws MachineLostException;

    public abstract boolean visitNStates(int n) throws CompositionException, MachineLostException;


    protected abstract void processNewAreas() throws RemoteException, ClientLostException;

    protected void processNewEntryPoints() throws RemoteException, ClientLostException {
        FlyState newEntryPoint;
        do {
            newEntryPoint = getNextEntryPoint();
            if (newEntryPoint != null) {
                while (!enqueNewEntryPoint(newEntryPoint)) {
                    processNewAreas();
                }

            }

        } while (newEntryPoint != null);

    }

    protected synchronized void waitForServer() {
        while (paused) {
            pausedAck = true;
            try {
                wait();
            } catch (InterruptedException iex) {
            }

        }

    }

    protected void getIddle() throws ClientLostException, RemoteException {
        stateOutputQueue.flush();
        TerminationInfo ti = new TerminationInfo(received(), sent(), visited());
        serverTask.iddleStation(client.clientId, ti);
        synchronized (this) {
            while (noWork()) {
                assert received() == stateInputQueue.getMovedToReadyQueue();
                try {
                    wait();
                } catch (InterruptedException iex) {
                }
                if (finishInvoked) return;
            }
        }
        serverTask.iddleStation(client.clientId, null);
    }

    protected abstract boolean noWork();
        
    protected void initArea(Area a) {
        a.setClientName("Client_" + client.clientId);

        String dumpFname = Preferences.getDumpFile();
        if (dumpFname != null) {
            a.setUserHandler(new FileDump(dumpFname + "_" + client.clientId));
        }

    }

    /**
     * Obtain a logger
     * @return logger
     */
    protected Logger logger() {
        return Logger.getLogger(
                "dchecker.ClientTask_" + Integer.toString(client.clientId));
    }
    /**
     * Set of unbound interfaces
     */
    protected Set<Integer> unbound;
    /**
     * RMI stub of the server object that takes care about this task
     */
    protected ServerTask serverTask;
    /**
     * Client that performs this task
     */
    protected RMIClientImpl client;
    protected int[] splitted;
    protected boolean end = false;
    protected boolean isIddle = false;

    protected abstract long visited();

    protected long sent() {
        return stateOutputQueue.getSent();
    }

    protected long received() {
        return stateInputQueue.getReceived();
    }

    protected synchronized FlyState getNextEntryPoint() {
        return stateInputQueue.getNextStateNonBlocking();
    }
    ClientInitData cid;

    /*********************************
     *  Methods Invoked through RMI  *
     *********************************/
    public synchronized void pause() throws RemoteException {
        paused = true;
        while (pausedAck != true) {
            try {
                wait();
            } catch (InterruptedException iex) {
            }

        }

        pausedAck = false;
    }

    public synchronized void resume() throws RemoteException {
        paused = false;
        notify();
    }

    public synchronized long finish() throws RemoteException {
        long visited = visited();
        logger().info("Task finished. Visited " + visited + " nodes.");
        end = true;
        finishInvoked = true;
        serverTask = null;
        notifyAll();
        return visited;
    }
    /***********************************
     *  Members used from RMI threads  *
     ***********************************/
    
    protected boolean finishInvoked = false;
    
    protected StateInputQueue stateInputQueue;
    protected StateOutputQueue stateOutputQueue;
    
    protected boolean paused = true;
    protected boolean pausedAck = false;
}
