/* $Id$
 *
 * dChecker - Distributed Behavior Protocols Checker 
 * Copyright (C) 2006-2007  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.ow2.dsrg.fm.dchecker.rmi;


import java.io.IOException;
import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;


import org.ow2.dsrg.fm.bptools.util.Pair;
import org.ow2.dsrg.fm.dchecker.distributed.Area;
import org.ow2.dsrg.fm.dchecker.distributed.spacemap.DistributionInfo;
import org.ow2.dsrg.fm.dchecker.errors.CompositionException;
import org.ow2.dsrg.fm.dchecker.fly.FlyFSM;
import org.ow2.dsrg.fm.dchecker.fly.FlyState;



/**
 * @author thorm
 */
public class RMIClientTaskImplDynamic extends RMIClientTaskImpl {
    
    LinkedList<Pair<int[],Integer>> updateMapRequests;
    
    
    public RMIClientTaskImplDynamic(RMIClientImpl client, ServerTask serverTask) throws RemoteException {
        super(client, serverTask);
        
        areas = new LinkedList<Area>();
        newAreas = new LinkedList<Area>();
        areaIterator = areas.iterator();
        
        reachedStatesFromUnassignedAreas = new HashSet<FlyState>();
        updateMapRequests = new LinkedList<Pair<int[],Integer > >();
        
        
    }
    
    protected boolean noWork(){
        return !stateInputQueue.nextStateAvailable() && 
                newAreas.isEmpty() &&
                updateMapRequests.isEmpty();
    }
    
    
    public synchronized boolean areaNotAssigned(){
        return areaCnt==0;
    }
    
    public boolean enqueNewEntryPoint(FlyState ep){
        automaton.getStateParts(ep, splitted);
        Area target = dInfo.get(splitted);
        if (target == null){
            return false;
        }
        target.newEntryPoint(ep);
        return true;
    }

    @Override
    public void init() throws IOException, ClientLostException{
        super.init();
    }
    
    
    private int areaCnt;
    
    public List<FlyState> getStackTrace(FlyState last) {
        
        automaton.getStateParts(last, splitted);
        Area target = dInfo.get(splitted);
        return target.getStackTrace(last);
        
    }
    
    public synchronized Area getSubtask() throws RemoteException {
        assert false : "Property splitting is not the same on server and client";
        return null;
    }
    
    
    
    
    protected long visited() {
        long ret = 0;
        for (Area a : areas) {
            ret += a.getNumberOfVisited();
        }
        return ret;
    }
    
    @Override
    public void finalCleanup() throws  NoSuchObjectException, IOException{
        super.finalCleanup();
        for (Area a : areas) {
            a.finishUserHandler();
        }
    }
    
    protected Iterator<Area> areaIterator;
    protected Area actualArea;
    
    protected void switchActualArea(){
        
        if (areaIterator.hasNext()){
            actualArea = areaIterator.next();
        } else {
            areaIterator = areas.iterator();
            if (areaIterator.hasNext()){
                actualArea = areaIterator.next();
            }
        }
    }
    
    
    
    public boolean visitNStates(int n) throws CompositionException, MachineLostException{
        boolean workDone = false;
        
        if (actualArea==null) {
            switchActualArea();
            if (actualArea==null) return false;
        }
        
        for (int i = 0; i< n ; i++){
            if (actualArea.traversal(this)){
                workDone = true;
            }else{
                switchActualArea();
            }
        }
        
        return workDone;
    }


    private void enqueReachedStates() throws ClientLostException{
        int[] parts = new int[automaton.getDimension()];
        Iterator<FlyState> it = reachedStatesFromUnassignedAreas.iterator();
        int cnt = 0;
        while (it.hasNext()){
            FlyState fs = it.next();
            automaton.getStateParts(fs,parts);
            Area a = dInfo.get(parts);
            if (a != null) {
                a.newEntryPoint(fs);
                it.remove();
                cnt++;
                continue;
            }
            Integer newTarget = cid.initialDI.get(parts);
            if (newTarget != null && newTarget!=-1 ){
                if (newTarget==client.clientId){
                    logger().info("Inconsistent initialDI and dInfo for " + fs.toString(automaton) );
                    continue;
                }
                stateOutputQueue.sendState(newTarget,fs);
                it.remove();
                cnt++;
            }
        }
        logger().info(cnt + " previously reached states assigned to areas. " +reachedStatesFromUnassignedAreas.size()+" left.");
    }

    /**
     * Automaton that is capable to generate state space on the fly
     */
    protected FlyFSM automaton;
    
    /**
     * list of areas assigned to this client
     */
    protected List<Area> areas;
    
    
    protected HashSet<FlyState> reachedStatesFromUnassignedAreas;
    
    /**
     * list of the new areas
     */
    private Queue<Area> newAreas;
    
    protected DistributionInfo<Area> dInfo;
    
    
    public static final int REACHED_BUT_NOT_ASSIGNED = -1;
    
    
    public void outOfArea(FlyState fs, int[] splitted) throws MachineLostException {
        
        Area target = dInfo.get(splitted);
        
        if (target != null) {
            target.newEntryPoint(fs);
        } else {
            Integer trg = cid.initialDI.get(splitted);
            if (trg == null){
                reachedStatesFromUnassignedAreas.add(fs);
                cid.initialDI.changeValueForWholeArea(splitted,REACHED_BUT_NOT_ASSIGNED);
                try{
                    serverTask.newAreaReached(fs);
                } catch (RemoteException re){
                    throw new ServerLostException();
                }
            }else if (trg==REACHED_BUT_NOT_ASSIGNED){
                reachedStatesFromUnassignedAreas.add(fs);
            }else{
                stateOutputQueue.sendState(trg,fs);
            }
        }
        
    }
    
    @Override
    protected long sent() {
        return stateOutputQueue.getSent()+reachedStatesFromUnassignedAreas.size();
    }
    
    
    @Override
    public  void processNewAreas() throws RemoteException, ClientLostException{
        if (automaton==null){
            automaton = serverTask.getWholeFlyFSM();
            int dim = automaton.getDimension();
            dInfo = new DistributionInfo(dim);
            splitted = new int[dim];
        }
        
        synchronized(this){
            for (Area a: newAreas){
                a.setFlyFSM(automaton);
                dInfo.addInterval(a,a);
                areas.add(a);
            }
            newAreas.clear();
            areaIterator = areas.iterator();
        }
        
        synchronized(updateMapRequests){
            boolean enque = !updateMapRequests.isEmpty();
            while (!updateMapRequests.isEmpty()){
                Pair<int[],Integer> rq = updateMapRequests.poll();
                int[] areaMember = rq.first;
                int clientId = rq.second;
                cid.initialDI.changeValueForWholeArea(areaMember,clientId);
                
            }
            if (enque) {
                enqueReachedStates();
            }
        }
        
    }
    
    
    
    /***********************************
     *  Methods Invoked through RMI    *
     ***********************************/
    
    /**
     * Add a new area. This should be invoked only if the station is iddle
     * @param a
     * @throws java.rmi.RemoteException
     */
    public synchronized void addArea(Area a) throws RemoteException {
        
        initArea(a);
        areaCnt++;
        
        newAreas.add(a);
        notifyAll();
    }
    
    public void updateAreaMap(int[] areaMember,int clientId) throws RemoteException{
        synchronized(updateMapRequests){
            updateMapRequests.add(new Pair<int[],Integer>(areaMember,clientId));
            
        }
    }
    
    
    /***********************************
     *  Members used from RMI threads  *
     ***********************************/
    
}
