/* $Id$
 *
 * dChecker - Distributed Behavior Protocols Checker 
 * Copyright (C) 2006-2007  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.ow2.dsrg.fm.dchecker.rmi;



import org.ow2.dsrg.fm.dchecker.distributed.ClientRepository;
import org.ow2.dsrg.fm.dchecker.distributed.ClientRepositoryDynamic;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.channels.ServerSocketChannel;
import java.rmi.ConnectException;
import java.rmi.Naming;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.server.UnicastRemoteObject;
import java.util.LinkedList;
import java.util.logging.Logger;
import org.ow2.dsrg.fm.bptools.Preferences;
import org.ow2.dsrg.fm.bptools.architecture.ParsedArchitecture;
import org.ow2.dsrg.fm.bptools.util.Pair;


/**
 * Implementation of server remote interface
 *
 * @author thorm
 */
public class RMIServerImpl extends UnicastRemoteObject implements Server {
    
    /**
     * Creates a new instance of ServerImpl.
     * Register the new instance in the rmi registry
     * @throws java.lang.Exception
     */
    public RMIServerImpl() throws Exception {
        super();
        register();
        
        tasksToDo = new LinkedList<Pair<UserInterface, ParsedArchitecture>>();
        
        
        if (Preferences.getSplitting()==Preferences.Splitting.Dynamic) {
            clients = new ClientRepositoryDynamic();
        } else {
            clients = new ClientRepository();
        }
        
    }
    
    /**
     * Register this instance in the rmi registry
     * @throws java.lang.Exception
     */
    private void register() throws Exception {
        LocateRegistry.createRegistry(1099);
        Naming.rebind("dcheckerServer", this);
        logger().info("bound in registry");
    }
    
    /**
     * This method is run by the main server thread.
     * The server registers the new clients and waits for the task from a console.
     * Then the task is received it creates RMIServerTaskImpl that contains the
     * information about the new task and invokes its method run.
     * When the checking of the task is finished it waits for the next task assignment.
     *
     * @throws java.lang.Exception
     */
    public void run() throws Exception {
        Pair<UserInterface, ParsedArchitecture> taskPair;
               
        while (!end) {
            synchronized (this) {
                while (tasksToDo.isEmpty()) {
                    logger().info("waiting for tasks");
                    wait();
                }
            }
            logger().info("task assigned");
            taskPair = tasksToDo.poll();
            
            boolean restart = true;
            
            while (restart) {                
                restart = false;
                try {
                    Preferences.Splitting splitting = Preferences.getSplitting();
                    
                    if (splitting == Preferences.Splitting.Hash) {
                        currentTask = new RMIServerTaskImplHash(this,
                                taskPair.first, taskPair.second);
                        logger().info("Splitting strategy : hash");
                    } else if (splitting == Preferences.Splitting.Dynamic) {
                        currentTask = new RMIServerTaskImplDynamic(this,
                                taskPair.first, taskPair.second);
                        logger().info("Splitting strategy : dynamic");
                    } else {
                        currentTask = new RMIServerTaskImplStatic(this,
                                taskPair.first, taskPair.second);
                        logger().info("Splitting strategy : static");
                    }
                    
                    currentTask.run();
                    
                    long visited = currentTask.finish();
                    
                    logger().info(
                            "Task finished. " + visited + " states visited, "
                            + currentTask.edgesBetweenStations
                            + " edges between machines");
                    taskPair.first.msg(
                            visited + " states visited. "
                            + currentTask.edgesBetweenStations
                            + " edges between machines");
                    
                    taskPair.first.finish();
                } catch (ConnectException connect) {
                    logger().info("Connection to console lost. Terminating task");
                    currentTask.finish();
                } catch (NoClientException e) {
                    logger().info("No client connected.");
                    taskPair.first.msg("No client connected.");
                    taskPair.first.finish();
                } catch (ClientLostException cl) {
                    //cl.printStackTrace();
                    logger().info("Client lost. Restarting the task");
                    taskPair.first.msg("Client lost. Restarting the task");
                    restart = true;
                    clients.unregisterClient(cl.clientId);
                    currentTask.finish();
                } /*catch (ParseException pe){
                    taskPair.first.msg("Exception :" + pe.toString());
                    taskPair.first.finish();
                } */catch (Exception e) {
                    e.printStackTrace();
                    taskPair.first.msg("Exception :" + e.toString());
                    currentTask.finish();
                    taskPair.first.finish();
                } finally {
                    clients.cleanClientTaskData();
                    currentTask = null;
                }
            }
        }
        
        ssc.close();
        
    }
    
    
    /**
     * Get the logger
     * @return logger
     */
    private Logger logger() {
        return Logger.getLogger("dchecker.Server");
    }
    
    /**
     * The end was signaled
     */
    boolean end;
    
    
    
    
    
    ServerSocketChannel ssc;
    
    /***********************************
     *  Methods Invoked through RMI    *
     ***********************************/
    
        /**
     * Register a new client
     * The new client id is assigned. From this moment the client should be
     * referenced by this identifier
     *
     * @param c Client RMI stub
     * @throws java.rmi.RemoteException
     * @return assigned identifier
     */
    public int registerClient(Client c,InetSocketAddress ia) throws RemoteException {
        
        int ret = clients.registerClient(c,ia);
        
        logger().info("Client registered");
        // if (currentTask!=null) currentTask.setClientTask(ret);
        return ret;
    }
    
    /**
     * Unregister client
     *
     * @param clientId client identifier
     * @throws java.rmi.RemoteException
     */
    public void unregisterClient(int clientId) throws RemoteException {
        
        try {
            clients.unregisterClient(clientId);
            logger().info("Client unregistered");
        } catch (IOException ioe) {
            ioe.printStackTrace();
        }
    }
    
    /**
     * Add a new task into the queue
     * The main thread that waits if there is no assigned task is notified.
     * The main thread takes the tasks from the queue, creates RMIServerTaskImpls
     * and runs the traversal
     *
     * @param ui the task was entered by this console
     * @param s the architecture description
     * @throws java.rmi.RemoteException
     */
    public synchronized void newTask(UserInterface ui, ParsedArchitecture s) throws RemoteException {
        ui.msg("Task accepted");
        if (currentTask != null) {
            ui.msg("Previous tasks not finished - architecture enqeued");
        }
        tasksToDo.add(new Pair<UserInterface, ParsedArchitecture>(ui, s));
        notifyAll();
    }

    
    
    
    /***********************************
     *  Members used from RMI threads  *
     ***********************************/

    /**
     * Registered clients
     */
    ClientRepository clients;
    
    /**
     * Queue of the tasks
     */
    LinkedList<Pair<UserInterface, ParsedArchitecture>> tasksToDo;
    
    /**
     * currently checked task
     */
    RMIServerTaskImpl currentTask;
}
