/* $Id$
 *
 * dChecker - Distributed Behavior Protocols Checker
 * Copyright (C) 2006-2007  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */

package org.ow2.dsrg.fm.dchecker.rmi;



import org.ow2.dsrg.fm.dchecker.errors.BadActivityException;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.util.Collections;
import java.util.Set;
import java.io.IOException;
import java.nio.channels.Selector;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.logging.Logger;
import org.ow2.dsrg.fm.bptools.BPEventTableImpl;
import org.ow2.dsrg.fm.bptools.Preferences;
import org.ow2.dsrg.fm.bptools.architecture.ParsedArchitecture;
import org.ow2.dsrg.fm.bptools.util.DOTPrintStream;
import org.ow2.dsrg.fm.bptools.util.Pair;
import org.ow2.dsrg.fm.dchecker.architecture.DChPreprocessedArchitecture;
import org.ow2.dsrg.fm.dchecker.distributed.Area;
import org.ow2.dsrg.fm.dchecker.distributed.ClientRepository;
import org.ow2.dsrg.fm.dchecker.distributed.spacemap.DistributionInfo;
import org.ow2.dsrg.fm.dchecker.errors.CompositionException;
import org.ow2.dsrg.fm.dchecker.fly.FlyFSM;
import org.ow2.dsrg.fm.dchecker.fly.FlyState;

/**
 * This class represents task on the server side
 * @author thorm
 */
public abstract class RMIServerTaskImpl extends UnicastRemoteObject implements ServerTask {
    
    TimeStamps timeStamps = new TimeStamps();
    
    public static final int QUEUE_SIZE = 1000000;
    
    /**
     * Creates a new instance of RMITaskImpl
     * Parse the architecture description transform it into the FlyFSM
     *
     * @param server server this task is running on
     * @param ui interface used by user to enter the task
     * @param architectureStr task textual description
     *
     * @throws java.rmi.RemoteException
     * @throws bpchecker.parser.javacc.ParseException syntax error
     */
    public RMIServerTaskImpl(RMIServerImpl server, UserInterface ui, ParsedArchitecture arch) throws Exception {
        super();
        timeStamps.start();
        this.server = server;
        
        outOfAreaStates = new LinkedList<FlyState>();
        newlyIddleStations = new LinkedList<Pair<Integer,TerminationInfo> >();
        iddleStations = new HashSet<Integer>();
        
        terminationDetectionMap = new HashMap<Integer,TerminationInfo>();
        
        architecture = new DChPreprocessedArchitecture(arch, new BPEventTableImpl());
        
        DOTPrintStream dumpNFSMStream = null;
        DOTPrintStream dumpDFSMStream = null;
        
        String dumpNFSM = Preferences.getDumpNFSM();
        if (dumpNFSM!=null){
            dumpNFSMStream = new DOTPrintStream(dumpNFSM);
        }
        
        String dumpDFSM = Preferences.getDumpDFSM();
        if (dumpDFSM!=null){
            dumpDFSMStream = new DOTPrintStream(dumpDFSM);
        }
        
        architecture.performTransformationChain(dumpNFSMStream,dumpDFSMStream);
        
        
        clientRep = server.clients;
        int machines = clientRep.size();
        
        if (machines == 0) {
            throw new NoClientException();
        }
        
        areas = createInitialAreas(machines);
        
        dInfo = createInitialDistributionInfo(clientRep.getClientIds());
        userInterface = ui;
        
        int idSizeInBits = architecture.getFlyfsm().identifierSize();
        int idSizeInIntegers;
        if (idSizeInBits%32>0){
            idSizeInIntegers = (idSizeInBits/32)+1;
        }else{
            idSizeInIntegers = idSizeInBits/32;
        }
        
        clientInitData = new ClientInitData(
                new InetSocketAddress(InetAddress.getLocalHost(), 8888),
                clientRep.createClientAddressMap(),
                dInfo,
                idSizeInIntegers*2,
                architecture.getUnbound()
                );
        
        timeStamps.areasPrepared();
        
        wakeUpper = new WakeUpper("Server Wakeupper", this);
        wakeUpper.start();
        
        
    }
    
    
    
    protected WakeUpper wakeUpper;
    
    protected abstract List<Area> createInitialAreas(int numberOfClients);
    protected abstract DistributionInfo<Integer> createInitialDistributionInfo(Set<Integer> clientIds );
    
    protected List<Area> areas;
    
    
    /**
     * Assign this task to the given client
     * @param clientId client identifier
     * @throws java.rmi.RemoteException
     */
    public boolean setClientTask(int clientId) throws ClientLostException,IOException {
        Client client = clientRep.getClient(clientId);
        
        if (client == null) {
            return false;
        }
        
        ClientTask ct;
        
        try {
            ct = client.setTask(this,splittingType);
        } catch (RemoteException e) {
            throw new ClientLostException(clientId);
        }
        
        clientRep.setClientTaskData(clientId, new ClientData(ct, clientId));
        
        
        return true;
    }
    
    /**
     * Assign this task to all clients
     * @throws java.rmi.RemoteException
     */
    public boolean setAllClientsTask() throws ClientLostException,IOException {
        boolean atLeastOne = false;
        
        for (int i = 0; i < clientRep.getMaxClientId(); i++) {
            atLeastOne = setClientTask(i) || atLeastOne;
        }
        
        if (atLeastOne){
            for (Area a:areas){
                Integer clientId = dInfo.get(a.getMinimalCoord());
                if (clientId !=null){
                    clientRep.addArea(clientId,a);
                }
            }
        }
        
        return atLeastOne;
    }
    
    
    
    /**
     * Get the RMI stub of client-side task representation for given client
     * @param clientId client id
     * @return client-side task represetnation
     *
     * synchronized ClientTask getClientTask(int clientId){
     * return clientRep.getClientTaskData(clientId).clientTask;
     * }
     */
    
    private synchronized boolean end() {
        return end;
    }
    
    public boolean shouldBeIddle(){
        boolean  emptyIddleList = false;
        synchronized (wakeUpper){
            emptyIddleList = newlyIddleStations.isEmpty();
        }
        return emptyIddleList && compositionException==null;
    }
    
    
    public boolean processEvents() throws ClientLostException{
        return processNewlyIddleStations();
    }
    
    public boolean processNewlyIddleStations() throws ClientLostException{
        Integer iddleStation;
        boolean newIddleStation;
        
        newIddleStation = false;
        while(true){
            Area a;
            
            Pair<Integer, TerminationInfo> p;
            synchronized(wakeUpper){
                p = newlyIddleStations.poll();
            }
            if (p==null) break;
            iddleStation = p.first;
            if (p.second!=null){
                terminationDetectionMap.put(iddleStation,p.second);
            }else{
                terminationDetectionMap.remove(iddleStation);
                continue;
            }
            
            a = getSomeWork(iddleStation);
            
            
            if (a == null) {
                logger().info("No work for " + iddleStation);
                iddleStations.add(iddleStation);
                newIddleStation = true;
            } else {
                assignArea(iddleStation,a);
            }
        }
        
        if (newIddleStation && (edgesBetweenStations=terminationDetection())>=0) return false;
        return true;
    }
    
    /**
     * This is run by the main thread.
     * In the loop waits for iddle client notifications and assigns new areas to
     * them
     *
     * @throws java.lang.Exception
     */
    public void run() throws ClientLostException,IOException,NoClientException {
        
        if (!setAllClientsTask()) {
            throw new NoClientException();
        }
        
        clientRep.resumeThemAll();
        timeStamps.clientsInitialized();
        synchronized (this){
            while (!end){
                
                while (shouldBeIddle()) {
                    try{wait();}catch (InterruptedException ed) {}
                }
                if (!processEvents()) break;
            }
        }
        if (compositionException != null) {
            reportError((CompositionException)compositionException);
        } else if (clientLostFromClient !=null){
            throw clientLostFromClient;
        } else{
            userInterface.msg("Checked.");
        }
        timeStamps.checkingComplete();
        wakeUpper.finish();
        logger().info(timeStamps.report());
    }
    
    
    protected synchronized long terminationDetection() throws ClientLostException{
        if (terminationDetectionMap.keySet().size()==clientRep.size()){
            long sent = 0;
            long received = 0;
            long visited = 0;
            for(Map.Entry<Integer,TerminationInfo> tie: terminationDetectionMap.entrySet()){
                TerminationInfo ti = tie.getValue();
                sent += ti.statesSent;
                received +=ti.statesReceived;
                visited += ti.statesVisited;
                System.out.println("Client "+tie.getKey()+": received-"+ti.statesReceived+" sent-"+ti.statesSent);
            }
            
            long waitingFor = sent - received;
            if (waitingFor==0){
                return sent;
            }
            
            if (waitingFor>0){
                logger().info("Pending messages "+ waitingFor + ". Visited "+visited +"states." );
            }
            return -1;
        }
        
        return -1;
        
    }
    
    
    private void reportError(CompositionException ce) throws RemoteException, ClientLostException {
        if (ce instanceof BadActivityException) {
            BadActivityException bae = (BadActivityException) ce;
            
            userInterface.msg(
                    "Event "
                    + architecture.getEventTable().getEventString(
                    bae.event));
        }
        
        String annotated = ce.toString() + "\n"
                + architecture.getAnnotatedProtocol(ce);
        
        userInterface.msg(annotated);
        List<FlyState> stackTrace = getStackTrace(ce);
        
        userInterface.msg(architecture.printCallStack(stackTrace));
        
    }
    
    private List<FlyState> getStackTrace(CompositionException compositionException) throws ClientLostException {
        LinkedList<FlyState> ret = new LinkedList<FlyState>();
        List<FlyState> stackPart;
        FlyState last = compositionException.f;
        
        FlyFSM ffsm = architecture.getFlyfsm();
        int[] parts = new int[ffsm.getDimension()];
        
        int clientId;
        
        while (last.parentData != null) {
            ffsm.getStateParts(last, parts);
            clientId = dInfo.get(parts);
            stackPart = clientRep.getStackTrace(clientId, last);
            last = stackPart.get(0);
            stackPart.remove(0);
            ret.addAll(0, stackPart);
        }
        
        ret.addFirst(last);
        
        return ret;
    }
    
    /**
     * Notify all clients that the checking of this task is finished
     * @throws java.rmi.RemoteException
     * @return number of states visited by all clients
     */
    public long finish() throws Exception {
        long ret = 0;
        
        ret = clientRep.finish();
        
        while (!unexportObject(this, false)) {
            ;
        }
        return ret;
    }
    
    
    protected boolean notAssignedAreaStateReached(FlyState s,int[]parts){
        assert false;
        return false;
    }
    
    
    
    /**
     * Get part of the architecture description - the automaton
     * @throws java.rmi.RemoteException
     * @return automaton describing the architecture
     */
    public ClientInitData getClientInitData() throws RemoteException {
        return clientInitData;
    }
    
    
    abstract protected Area getSomeWork(int targetClientId) throws  ClientLostException;
    
    protected void processReachedArea() throws ClientLostException {}
    
    protected void assignArea(int clientId, Area a)throws ClientLostException{
        logger().info("Area " + a + " assigned to " + clientId);
        synchronized(this){
            dInfo.addInterval(a, clientId);
            terminationDetectionMap.remove(clientId);
        }
        clientRep.addArea(clientId, a);
    }
    
    
    
    
    /**
     * Obtain a logger
     * @return logger
     */
    protected Logger logger() {
        return Logger.getLogger("dchecker.ServerTask");
    }
    
    
    protected static int indexOfMaximum(int[]a) {
        if (a.length == 0) {
            return -1;
        }
        int maxValue = a[0];
        int maxIndex = 0;
        
        for (int i = 0; i < a.length; i++) {
            if (maxValue < a[i]) {
                maxValue = a[i];
                maxIndex = i;
            }
        }
        return maxIndex;
    }
    
    protected static int[] sortDimsBySize(int[] a){
        ArrayList<Pair<Integer,Integer>> tmp = new ArrayList<Pair<Integer,Integer>>();
        for (int i = 0; i<a.length; i++){
            tmp.add(new Pair<Integer,Integer>(i,a[i]));
        }
        
        Collections.sort(tmp,new Comparator<Pair<Integer,Integer>>(){
            public int compare(Pair<Integer,Integer> l,Pair<Integer,Integer> r){
                return l.second-r.second;
            }
        }
        );
        
        int[] ret = new int [a.length];
        for ( int i = 0; i<a.length ; i++){
            ret[i] = tmp.get(i).first;
        }
        
        return ret;
    }
    
    
    private LinkedList<FlyState> outOfAreaStates;
    private LinkedList<Pair<Integer, TerminationInfo> > newlyIddleStations;
    protected HashSet<Integer> iddleStations;
    
    protected HashMap<Integer,TerminationInfo> terminationDetectionMap;
    
    /**
     * Checking is done
     * Error was found or all clients are iddle
     */
    boolean end = false;
    
    /**
     * Used to distinguish the first assinment of the task
     */
    boolean taskStart = true;
    
    /**
     * Keeps the data about state space distribution
     */
    DistributionInfo<Integer> dInfo;
    
    /**
     * Architecture to be checked
     */
    DChPreprocessedArchitecture architecture;
    
    /**
     * interface used by user to enter the current task
     */
    UserInterface userInterface;
    
    /**
     * List of client-side task representations
     */
// ArrayList<ClientData> clientTasks;
    
    
    /**
     * this server
     */
    RMIServerImpl server;
    
    /**
     * Maintains the registered clients
     */
    ClientRepository clientRep;
    
    long edgesBetweenStations = 0;
    
    private ClientInitData clientInitData;
    
    private CompositionException compositionException;
    private ClientLostException clientLostFromClient;
    
    
    
    Selector selector;
    Preferences.Splitting splittingType;
    
    
    public class ClientData {
        public ClientData(ClientTask ct, int i) throws IOException {
            clientTask = ct;
        }
        
        public long finish() throws RemoteException,IOException{
            long ret = clientTask.finish();
            //closeSockets();
            return ret;
        }
        
        public ClientTask clientTask;
        
    }
    
    
    
    /***********************************
     *  Methods Invoked through RMI    *
     ***********************************/
    
    
    /**
     * Used by clients to notify the server that they are iddle
     * @param clientId iddle client identifier
     * @throws java.rmi.RemoteException
     */
    public void iddleStation(int clientId, TerminationInfo ti) throws RemoteException {
        synchronized (wakeUpper){
            newlyIddleStations.add(new Pair<Integer, TerminationInfo>(clientId, ti));
            wakeUpper.notify();
        }
        
    }
    
    /**
     * Client uses this method to report composition error found in the
     * architecture
     *
     * @param s exception that keeps the error description
     * @throws java.rmi.RemoteException
     */
    public synchronized void errorFound(Exception s) throws RemoteException {
        if (s instanceof ClientLostException){
            clientLostFromClient = (ClientLostException)s;
        }else if (s instanceof CompositionException){
            compositionException = (CompositionException)s;
        }else{
            assert false;
        }
        end = true;
        
        notify();
    }
    
    
    public FlyFSM getWholeFlyFSM() throws RemoteException {
        return architecture.getFlyfsm();
    }
    
    /**
     * Get part of the architecture description - list of unbound interfaces
     * @throws java.rmi.RemoteException
     * @return
     */
    public Set<Integer> getUnbound()throws RemoteException {
        return architecture.getUnbound();
    }
    
    
    /***********************************
     *  Members used from RMI threads  *
     ***********************************/
    
    public DChPreprocessedArchitecture getArchitecture() {
        return architecture;
    }
    
}

