/* $Id$
 *
 * dChecker - Distributed Behavior Protocols Checker 
 * Copyright (C) 2006-2007  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.fm.dchecker.rmi;

import org.ow2.dsrg.fm.dchecker.fly.FlyState;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.nio.BufferUnderflowException;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.SelectionKey;
import java.nio.channels.Selector;
import java.nio.channels.ServerSocketChannel;
import java.nio.channels.SocketChannel;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.Set;

/**
 *
 * @author tom
 */
public class StateInputQueue extends ClientStateQueue {

    protected ServerSocket ss = null;
    List<NIOReadBuffer> inputChannels;
    ReadingLoop readingLoop;
    protected int received;
    final Object toWakeUp;

    public StateInputQueue(ClientInitData cid, int portToBind, int clientId, final Object toWakeup) {
        super(cid, portToBind, clientId);
        this.toWakeUp = toWakeup;
    }

    public void bind() throws IOException {

        ServerSocketChannel ssc = ServerSocketChannel.open();
        ss = ssc.socket();

        InetSocketAddress address = new InetSocketAddress(ownPort);
        System.out.println("Binding to " + address);
        ss.bind(address);

    }

    public void acceptConnectionsFromOtherClients() throws IOException {
        inputChannels = new ArrayList<NIOReadBuffer>(otherClients.size());
        // accept connections from other clients
        for (int i = 0; i < otherClients.size() - 1; i++) {
            SocketChannel sch = ss.accept().getChannel();
            sch.configureBlocking(false);
            NIOReadBuffer niob = new NIOReadBuffer(sch, stateSizeInInts);
            inputChannels.add(niob);
        }
    }

    public void close() throws IOException {
        readingLoop.terminate();
        ss.close();
    }

    public boolean nextStateAvailable() {
        return readingLoop.peek() != null;
    }

    public FlyState getNextStateNonBlocking() {
        FlyState fs = readingLoop.poll();
        if (fs != null) {
            received++;
        }
        return fs;
    }

    public void start() throws IOException {
        readingLoop = new ReadingLoop(inputChannels, toWakeUp);
        Thread t = new Thread(readingLoop, "Client_" + clientId + "_Selector");
        t.start();
    }

    public int getReceived() {
        return received;
    }
    
    public int getMovedToReadyQueue(){
        return readingLoop.getStatesMovedToReadyQueue();
    }
}

class ReadingLoop implements Runnable {

    Selector selector;
    Queue<FlyState> readyStates = new LinkedList<FlyState>();
    private int statesMovedToReadyQueue = 0;
    boolean running = true;
    final Object m_toWakeUp;
    List<NIOReadBuffer> inputChannels;

    public ReadingLoop(List<NIOReadBuffer> inputChannels, final Object toWakeUp) throws IOException {
        this.inputChannels = inputChannels;
        m_toWakeUp = toWakeUp;
        synchronized (Selector.class) {
            selector = Selector.open();
        }
        for (NIOReadBuffer rb : inputChannels) {
            rb.register(selector);
        }
    }

    //called from synchronized block
    public void moveAllStatesToQueue(NIOReadBuffer rb) {
        FlyState fs = null;
        while ((fs = rb.readState()) != null) {
            readyStates.add(fs);
            statesMovedToReadyQueue++;
        }
    }

    public synchronized FlyState peek() {
        return readyStates.peek();
    }

    public synchronized FlyState poll() {
        return readyStates.poll();
    }

    public synchronized void terminate() {
        running = false;
        selector.wakeup();
    }

    public void run() {
        try {
            while (running) {
                int selectedSize = selector.select();
                if (selectedSize > 0) {
                    Set<SelectionKey> selectedKeys = selector.selectedKeys();
                    Iterator<SelectionKey> it = selectedKeys.iterator();
                    while (it.hasNext()) {
                        SelectionKey sk = it.next();
                        NIOReadBuffer niob = (NIOReadBuffer) sk.attachment();
                        niob.fetchDataIntoBuffer();
                        it.remove();
                        synchronized (m_toWakeUp) {
                            synchronized (this) {
                                moveAllStatesToQueue(niob);
                                m_toWakeUp.notify();
                            }
                        }
                    }
                }
            }

            System.out.println("Closing input sockets");
            for (NIOReadBuffer sc : inputChannels) {
                sc.close();
            }
            System.out.println("Closing selector");
            selector.close();

        } catch (IOException ioe) {
            ioe.printStackTrace();
            assert false;
        }
    }

    public synchronized int getStatesMovedToReadyQueue() {
        return statesMovedToReadyQueue;
    }
    }

class NIOReadBuffer {

    int[] data;
    int position;
    ByteBuffer buffer;
    SocketChannel channel;
    final int stateSizeInInts;

    public NIOReadBuffer(SocketChannel c, int _stateSizeInInts) {
        channel = c;
        stateSizeInInts = _stateSizeInInts;
        buffer = ByteBuffer.allocateDirect(stateSizeInInts * 4 * ClientStateQueue.BufferSizeInStates);
        buffer.clear();
        buffer.flip();
        data = new int[stateSizeInInts];
        position = 0;
    }

    public synchronized FlyState readState() {
        for (; position < stateSizeInInts; position++) {
            try {
                data[position] = buffer.getInt();
            } catch (BufferUnderflowException bue) {
                assert buffer.remaining() == 0;
                return null;
            }
        }
        position = 0;
        FlyState ret = new FlyState(stateSizeInInts * 8 * 4 / 2);
        ret.readData(data);
        return ret;
    }

    public void register(Selector s) throws ClosedChannelException {
        channel.register(s, SelectionKey.OP_READ, this);
    }

    void close() throws IOException {
        channel.close();
    }

    synchronized void fetchDataIntoBuffer() throws IOException {
        int remaining = buffer.remaining();
        assert remaining == 0;
        buffer.clear();
        channel.read(buffer);
        buffer.flip();
    }
    }
