/* $Id$
 *
 * dChecker - Distributed Behavior Protocols Checker 
 * Copyright (C) 2006-2007  DSRG, Charles University in Prague
 *                          http://dsrg.mff.cuni.cz/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 *
 */
package org.ow2.dsrg.fm.dchecker.rmi;

import org.ow2.dsrg.fm.dchecker.fly.FlyState;
import java.io.IOException;
import java.net.InetSocketAddress;
import java.nio.BufferOverflowException;
import java.nio.ByteBuffer;
import java.nio.channels.SocketChannel;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

/**
 *
 * @author tom
 */
public class StateOutputQueue extends ClientStateQueue {

    Map<Integer, OutputChanBufPair> outputChannels = new HashMap<Integer, OutputChanBufPair>();
    protected int sent;

    public StateOutputQueue(ClientInitData cid, int portToBind, int clientId) {
        super(cid, portToBind, clientId);
    }

    public void connectToOtherClients() throws ClientLostException {
        for (Integer i : otherClients.keySet()) {

            //do not connect to itself
            if (clientId == i) {
                continue;
            }

            int attempts = 0;
            while (true) {
                try {
                    InetSocketAddress isa = otherClients.get(i);
                    assert isa != null;
                    OutputChanBufPair ocbp = new OutputChanBufPair(i, stateSizeInInts, clientId, isa);
                    outputChannels.put(i, ocbp);
                    System.out.println("OK");
                    break;
                } catch (IOException e) {
                    System.out.println("Failed");
                    attempts++;
                    if (attempts == 10) {
                        throw new ClientLostException(i);
                    }
                    synchronized (this) {
                        try {
                            wait(500);
                        } catch (InterruptedException ie) {

                        }
                    }
                }
            }
        }
    }

    public void sendState(int client, FlyState fs) throws ClientLostException {
        OutputChanBufPair out = outputChannels.get(client);
        out.sendState(fs);
        sent++;
    }

    public void flush() throws ClientLostException {
        for (Map.Entry<Integer, OutputChanBufPair> e : outputChannels.entrySet()) {
            OutputChanBufPair v = e.getValue();
            v.flush();
        }
    }

    public int getSent() {
        return sent;
    }

    public void close() throws IOException {

        for (OutputChanBufPair ocbp : outputChannels.values()) {
            ocbp.terminate();
        }
    }
}

class OutputChanBufPair implements Runnable {

    SocketChannel ch;
    ByteBuffer buf;
    int[] oneStateBuffer;
    LinkedList<FlyState> queue;
    boolean flushBit = false;
    boolean running = true;
    int stateSizeInInts;
    int clientId;
    IOException e = null;
    int sent = 0;
    int queuedStates = 0;            

    public OutputChanBufPair(int target, int _stateSizeInInts, int clientId, InetSocketAddress isa) throws IOException {
        System.out.print("Connecting to " + isa + "... ");
        ch = SocketChannel.open();
        ch.connect(isa);
        ch.configureBlocking(true);
        this.clientId = clientId;
        stateSizeInInts = _stateSizeInInts;
        buf = ByteBuffer.allocateDirect(stateSizeInInts * 4 * ClientStateQueue.BufferSizeInStates);
        oneStateBuffer = new int[stateSizeInInts];
        queue = new LinkedList<FlyState>();
        new Thread(this, "Client_" + clientId + " to " + target + " writer").start();
    }

    public synchronized void flush() throws ClientLostException {
        flushBit = true;
        notifyAll();
        if (e != null) {
            throw new ClientLostException(clientId);
        }
    }

    public synchronized void sendState(FlyState fs) throws ClientLostException {
        queue.add(fs);
        queuedStates++;
        notifyAll();
        if (e != null) {
            throw new ClientLostException(clientId);
        }
    }

    public synchronized void terminate() {
        running = false;
        notify();
    }

    public synchronized void run() {
        try {
            while (running) {
                FlyState fs;
                while ((fs = queue.poll()) != null) {
                    writeState(fs);
                }
                if (flushBit) {
                    flushImpl();
                    assert queuedStates == sent;
                    flushBit = false;
                }
                try {                    
                    wait();
                } catch (InterruptedException ex) {
                }
            }
            ch.close();

        } catch (IOException ex) {
            e = ex;
        }
    }

    private void writeState(FlyState fs) throws IOException {
        fs.writeData(oneStateBuffer);
        for (int i = 0; i < stateSizeInInts;) {
            try {
                buf.putInt(oneStateBuffer[i]);
                i++;
            } catch (BufferOverflowException bue) {
                buf.flip();
                int bytesWritten = ch.write(buf);
                sent += bytesWritten/(stateSizeInInts*4);
                buf.clear();
            }
        }
    }

    private boolean flushImpl() throws IOException {
        buf.flip();
        if (buf.hasRemaining()) {
            int bytesWritten = ch.write(buf);
            sent += bytesWritten/(stateSizeInInts*4);
        }
        buf.clear();
        return true;
    }
}
